#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Commandline Utilities for Managing the IDAES Data Directory"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: protected access issues
# pylint: disable=protected-access

__author__ = "John Eslick"

import os
import click
import idaes
from idaes.commands import cb


@cb.command(name="data-directory", help="Show IDAES data directory")
@click.option("--exists", is_flag=True, help="Show if the directory exists")
@click.option("--create", is_flag=True, help="Create the directory")
def data_directory(exists, create):
    if create:
        click.echo("Creating {}".format(idaes.data_directory))
        idaes._create_data_dir()
    elif exists:
        click.echo(os.path.exists(idaes.data_directory))
    else:
        click.echo(idaes.data_directory)


@cb.command(name="bin-directory", help="Show IDAES executable file directory")
@click.option("--exists", is_flag=True, help="Show if the directory exists")
@click.option("--create", is_flag=True, help="Create the directory")
def bin_directory(exists, create):
    if create:
        click.echo("Creating {}".format(idaes.bin_directory))
        idaes._create_bin_dir()
    elif exists:
        click.echo(os.path.exists(idaes.bin_directory))
    else:
        click.echo(idaes.bin_directory)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains tests for the variable replace transformation.
"""

import pytest
import pyomo.environ as pyo
import idaes.core.plugins

__author__ = "John Eslick"


@pytest.mark.unit
def test_1():
    # Test scalar variables
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.x = pyo.Var(initialize=2)
    m.y = pyo.Var(initialize=3)
    m.z = pyo.Var(initialize=0)
    m.c1 = pyo.Constraint(expr=m.z == m.x + m.y)
    m.e1 = pyo.Expression(expr=m.x**m.y)
    m.o1 = pyo.Objective(expr=m.y - m.x)

    assert m.c1.body() == -5  # hope constraint arrangement is deterministic
    assert pyo.value(m.e1) == 8
    assert pyo.value(m.o1) == 1
    rp.apply_to(m, substitute=[(m.y, m.x)])
    assert m.c1.body() == -4
    assert pyo.value(m.e1) == 4
    assert pyo.value(m.o1) == 0


@pytest.mark.unit
def test_2():
    # Test vector variables and sums
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.x = pyo.Var(["a", "b", "c"], initialize=2)
    m.y = pyo.Var(initialize=3)
    m.z = pyo.Var(initialize=0)
    m.c1 = pyo.Constraint(expr=m.z == m.x["a"] + m.x["b"] + m.x["c"])
    m.e1 = pyo.Expression(expr=sum(m.x[i] for i in m.x))

    assert m.c1.body() == -6  # hope constraint arrangement is deterministic
    assert pyo.value(m.e1) == 6
    rp.apply_to(m, substitute=[(m.x["c"], m.y)])
    assert m.c1.body() == -7
    assert pyo.value(m.e1) == 7


@pytest.mark.unit
def test_3():
    # Test expression in constraint
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.x = pyo.Var(["a", "b", "c"], initialize=2)
    m.y = pyo.Var(initialize=3)
    m.z = pyo.Var(initialize=0)
    m.e1 = pyo.Expression(expr=sum(m.x[i] for i in m.x))
    m.c1 = pyo.Constraint(expr=m.z == m.e1)

    assert m.c1.body() == -6  # hope constraint arrangement is deterministic
    rp.apply_to(m, substitute=[(m.x["c"], m.y)])
    assert m.c1.body() == -7


@pytest.mark.unit
def test_4():
    # Test expression in objective
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.x = pyo.Var(["a", "b", "c"], initialize=2)
    m.y = pyo.Var(initialize=3)
    m.z = pyo.Var(initialize=0)
    m.e1 = pyo.Expression(expr=sum(m.x[i] for i in m.x))
    m.c1 = pyo.Constraint(expr=m.z == m.e1)
    m.o1 = pyo.Objective(expr=m.e1)

    assert pyo.value(m.o1) == 6
    rp.apply_to(m, substitute=[(m.x["c"], m.y)])
    assert pyo.value(m.o1) == 7


@pytest.mark.unit
def test_4():
    # Test in a hierarchical model
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.b1 = pyo.Block()
    m.b1.b2 = pyo.Block()
    x = m.b1.b2.x = pyo.Var(["a", "b", "c"], initialize=2)
    m.y = pyo.Var(initialize=3)
    m.z = pyo.Var(initialize=0)
    m.e1 = pyo.Expression(expr=sum(x[i] for i in x))
    m.b1.c1 = pyo.Constraint(expr=m.z == m.e1)
    m.o1 = pyo.Objective(expr=m.e1)

    assert pyo.value(m.o1) == 6
    assert m.b1.c1.body() == -6
    assert pyo.value(m.e1) == 6
    rp.apply_to(m, substitute=[(x["c"], m.y)])
    assert pyo.value(m.o1) == 7
    assert m.b1.c1.body() == -7
    assert pyo.value(m.e1) == 7


@pytest.mark.unit
def test_5():
    # Test indexed var replace
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.b1 = pyo.Block()
    m.b1.b2 = pyo.Block()
    x = m.b1.b2.x = pyo.Var(["a", "b", "c"], [1, 2, 3], initialize=2)
    m.y = pyo.Var(["a", "b", "c", "d"], [1, 2, 3], initialize=3)
    m.z = pyo.Var(initialize=0)
    m.e1 = pyo.Expression(expr=sum(x[i] for i in x))
    m.b1.c1 = pyo.Constraint(expr=m.z == m.e1)
    m.o1 = pyo.Objective(expr=m.e1)

    assert pyo.value(m.o1) == 18
    assert m.b1.c1.body() == -18
    assert pyo.value(m.e1) == 18
    rp.apply_to(m, substitute=[(x, m.y)])
    assert pyo.value(m.o1) == 27
    assert m.b1.c1.body() == -27
    assert pyo.value(m.e1) == 27


@pytest.mark.unit
def test_6():
    # Test non-variable exception
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.b1 = pyo.Block()
    m.b1.b2 = pyo.Block()
    x = m.b1.b2.x = pyo.Var(["a", "b", "c"], [1, 2, 3], initialize=2)
    m.y = pyo.Var(["a", "b", "c", "d"], [1, 2, 3], initialize=3)
    m.z = pyo.Var(initialize=0)

    with pytest.raises(TypeError):
        rp.apply_to(m, substitute=[(x, m.b1)])
    with pytest.raises(TypeError):
        rp.apply_to(m, substitute=[(m.b1, x)])


@pytest.mark.unit
def test_7():
    # Test replace indexed by non-indexed
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.b1 = pyo.Block()
    m.b1.b2 = pyo.Block()
    x = m.b1.b2.x = pyo.Var(["a", "b", "c"], [1, 2, 3], initialize=2)
    m.y = pyo.Var(["a", "b", "c", "d"], [1, 2, 3], initialize=3)
    m.z = pyo.Var(initialize=0)

    assert x.is_indexed()
    assert not m.z.is_indexed()
    with pytest.raises(TypeError):
        rp.apply_to(m, substitute=[(x, m.z)])


@pytest.mark.unit
def test_8():
    # Test replace indexed by indexed var that doesn't have enough/right indexes
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.b1 = pyo.Block()
    m.b1.b2 = pyo.Block()
    x = m.b1.b2.x = pyo.Var(["a", "b", "c"], [1, 2, 3], initialize=2)
    m.y = pyo.Var(["a", "b", "d"], [1, 2, 3], initialize=3)

    with pytest.raises(ValueError):
        rp.apply_to(m, substitute=[(x, m.y)])


@pytest.mark.unit
def test_9():
    # test with references the way we handle time indexing a lot in IDAES
    rp = pyo.TransformationFactory("replace_variables")
    block_set = {1, 2, 3}
    m = pyo.ConcreteModel()
    m.b1 = pyo.Block(block_set)
    for i in block_set:
        m.b1[i].x = pyo.Var(initialize=2)
    m.y = pyo.Var([1, 2, 3], initialize=3)
    m.xx = pyo.Reference(m.b1[:].x)
    m.display()

    m.e1 = pyo.Expression(expr=sum(m.xx[i] for i in m.xx))
    m.e2 = pyo.Expression(expr=sum(m.b1[i].x for i in m.b1))

    assert pyo.value(m.e1) == 6
    assert pyo.value(m.e2) == 6
    rp.apply_to(m, substitute=[(m.xx, m.y)])
    assert pyo.value(m.e1) == 9
    assert pyo.value(m.e2) == 9


@pytest.mark.unit
def test_10():
    # test with more than one variable in the list
    rp = pyo.TransformationFactory("replace_variables")
    m = pyo.ConcreteModel()
    m.x = pyo.Var(["a", "b", "c"], initialize=2)
    m.a = pyo.Var(initialize=5)
    m.b = pyo.Var(initialize=6)
    m.c = pyo.Var(initialize=7)

    m.e1 = pyo.Expression(expr=sum(m.x[i] for i in m.x))

    assert pyo.value(m.e1) == 6
    rp.apply_to(m, substitute=[(m.x["a"], m.a), (m.x["b"], m.b), (m.x["c"], m.c)])
    assert pyo.value(m.e1) == 18



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Thickener unit model.

Unit model is derived from:

R. Burger, F. Concha, K.H. Karlsen, A. Narvaez,
Numerical simulation of clarifier-thickener units treating ideal
suspensions with a flux density function having two inflection points,
Mathematical and Computer Modelling 44 (2006) 255–275
doi:10.1016/j.mcm.2005.11.008

Settling velocity function from:

N.G. Barton, C.-H. Li, S.J. Spencer, Control of a surface of discontinuity in continuous thickeners,
Journal of the Australian Mathematical Society Series B 33 (1992) 269–289

"""
# Import Python libraries
import logging
from pandas import DataFrame

# Import Pyomo libraries
from pyomo.environ import Expr_if, inequality, units, Var
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Port

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.models.unit_models.separator import (
    Separator,
    SplittingType,
    EnergySplittingType,
)
from idaes.core.initialization import BlockTriangularizationInitializer
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.units_of_measurement import report_quantity
from idaes.core.util.constants import Constants as CONST


__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger("idaes.unit_model")


@declare_process_block_class("Thickener0D")
class Thickener0DData(UnitModelBlockData):
    """
    Thickener0D Unit Model Class
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. Flash units do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Flash units do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "energy_split_basis",
        ConfigValue(
            default=EnergySplittingType.equal_temperature,
            domain=EnergySplittingType,
            description="Type of constraint to write for energy splitting",
            doc="""Argument indicating basis to use for splitting energy this is
    not used for when ideal_separation == True.
    **default** - EnergySplittingType.equal_temperature.
    **Valid values:** {
    **EnergySplittingType.equal_temperature** - outlet temperatures equal inlet
    **EnergySplittingType.equal_molar_enthalpy** - outlet molar enthalpies equal
    inlet,
    **EnergySplittingType.enthalpy_split** - apply split fractions to enthalpy
    flows.}""",
        ),
    )
    CONFIG.declare(
        "solid_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for solid phase",
            doc="""Property parameter object used to define solid phase property
    calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "solid_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing solid phase property packages",
            doc="""A ConfigBlock with arguments to be passed to a solid phase property
    block(s) and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "liquid_property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for liquid phase",
            doc="""Property parameter object used to define liquid phase property
        calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "liquid_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing liquid phase property packages",
            doc="""A ConfigBlock with arguments to be passed to a liquid phase property
        block(s) and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}""",
        ),
    )

    default_initializer = BlockTriangularizationInitializer

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call super().build to setup dynamics
        super().build()

        # Build Solid Phase
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.solid_property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.solid_inlet_state = self.config.solid_property_package.build_state_block(
            self.flowsheet().time, doc="Solid properties in separator", **tmp_dict
        )

        # Add solid splitter
        self.solid_split = Separator(
            property_package=self.config.solid_property_package,
            property_package_args=self.config.solid_property_package_args,
            outlet_list=["underflow", "overflow"],
            split_basis=SplittingType.totalFlow,
            ideal_separation=False,
            mixed_state_block=self.solid_inlet_state,
            has_phase_equilibrium=False,
            material_balance_type=self.config.material_balance_type,
            momentum_balance_type=self.config.momentum_balance_type,
            energy_split_basis=self.config.energy_split_basis,
        )

        # Add solid ports
        self.add_port(
            name="solid_inlet",
            block=self.solid_inlet_state,
            doc="Solid inlet to thickener",
        )
        self.solid_underflow = Port(extends=self.solid_split.underflow)
        self.solid_overflow = Port(extends=self.solid_split.overflow)

        # Build liquid Phase
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.liquid_property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.liquid_inlet_state = self.config.liquid_property_package.build_state_block(
            self.flowsheet().time, doc="liquid properties in separator", **tmp_dict
        )

        # Add liquid splitter
        self.liquid_split = Separator(
            property_package=self.config.liquid_property_package,
            property_package_args=self.config.liquid_property_package_args,
            outlet_list=["underflow", "overflow"],
            split_basis=SplittingType.totalFlow,
            ideal_separation=False,
            mixed_state_block=self.liquid_inlet_state,
            has_phase_equilibrium=False,
            material_balance_type=self.config.material_balance_type,
            momentum_balance_type=self.config.momentum_balance_type,
            energy_split_basis=self.config.energy_split_basis,
        )

        # Add liquid ports
        self.add_port(
            name="liquid_inlet",
            block=self.liquid_inlet_state,
            doc="liquid inlet to thickener",
        )
        self.liquid_underflow = Port(extends=self.liquid_split.underflow)
        self.liquid_overflow = Port(extends=self.liquid_split.overflow)

        # Add additional variables and constraints
        uom = self.solid_inlet_state.params.get_metadata().derived_units

        self.area = Var(
            initialize=1,
            units=uom.AREA,
            doc="Cross sectional area of thickener",
        )

        # Volumetric Flowrates
        self.flow_vol_feed = Var(
            self.flowsheet().time,
            initialize=0.7,
            units=uom.FLOW_VOL,
            bounds=(0, None),
            doc="Total volumetric flowrate of feed",
        )
        self.flow_vol_overflow = Var(
            self.flowsheet().time,
            initialize=0.7,
            units=uom.FLOW_VOL,
            bounds=(0, None),
            doc="Total volumetric flowrate of overflow",
        )
        self.flow_vol_underflow = Var(
            self.flowsheet().time,
            initialize=0.7,
            units=uom.FLOW_VOL,
            bounds=(0, None),
            doc="Total volumetric flowrate of underflow",
        )

        # Solid Fractions
        self.solid_fraction_feed = Var(
            self.flowsheet().time,
            initialize=0.7,
            units=units.dimensionless,
            bounds=(0, None),
            doc="Volume fraction of solids in feed",
        )
        self.solid_fraction_underflow = Var(
            self.flowsheet().time,
            initialize=0.7,
            units=units.dimensionless,
            bounds=(0, None),
            doc="Volume fraction of solids in underflow",
        )
        self.solid_fraction_overflow = Var(
            self.flowsheet().time,
            initialize=0.7,
            units=units.dimensionless,
            bounds=(0, None),
            doc="Volume fraction of solids in overflow",
        )

        # Flux densities
        self.flux_density_underflow = Var(
            self.flowsheet().time,
            initialize=0,
            units=uom.VELOCITY,
            doc="Kynch flux density in underflow",
        )
        self.flux_density_overflow = Var(
            self.flowsheet().time,
            initialize=0,
            units=uom.VELOCITY,
            doc="Kynch flux density in overflow",
        )

        # Parameters
        self.particle_size = Var(
            self.flowsheet().time,
            initialize=1e-5,
            units=uom.LENGTH,
            doc="Characteristic length of particle",
        )
        self.v0 = Var(
            self.flowsheet().time,
            initialize=1e-4,
            units=uom.VELOCITY,
            doc="Stokes velocity of individual particle",
        )
        self.v1 = Var(
            initialize=1e-5,
            units=uom.VELOCITY,
            doc="Superficial velocity of a Darcy type flow of liquid",
        )
        self.C = Var(
            initialize=5,
            units=units.dimensionless,
            bounds=(0, None),
            doc="Settling velocity exponent",
        )
        self.solid_fraction_max = Var(
            initialize=0.9,
            units=units.dimensionless,
            bounds=(0, 1),
            doc="Maximum achievable solids volume fraction",
        )

        # ---------------------------------------------------------------------------------------------
        # Constraints
        @self.Constraint(self.flowsheet().time)
        def feed_flowrate(b, t):
            return b.flow_vol_feed[t] == (
                b.solid_inlet_state[t].flow_vol
                + units.convert(b.liquid_inlet_state[t].flow_vol, to_units=uom.FLOW_VOL)
            )

        @self.Constraint(self.flowsheet().time)
        def overflow_flowrate(b, t):
            return b.flow_vol_overflow[t] == (
                b.solid_split.overflow_state[t].flow_vol
                + units.convert(
                    b.liquid_split.overflow_state[t].flow_vol, to_units=uom.FLOW_VOL
                )
            )

        @self.Constraint(self.flowsheet().time)
        def underflow_flowrate(b, t):
            return b.flow_vol_underflow[t] == (
                b.solid_split.underflow_state[t].flow_vol
                + units.convert(
                    b.liquid_split.underflow_state[t].flow_vol, to_units=uom.FLOW_VOL
                )
            )

        # Eqn 2.8 from [1]
        @self.Constraint(self.flowsheet().time)
        def flux_density_function_overflow(b, t):
            u = b.solid_fraction_overflow
            return b.flux_density_overflow[t] == Expr_if(
                IF=inequality(0, u[t], b.solid_fraction_max),
                THEN=b.v0[t] * u[t] * (1 - u[t] / b.solid_fraction_max) ** b.C
                + b.v1 * u[t] ** 2 * (b.solid_fraction_max - u[t]),
                ELSE=0 * units.m * units.s**-1,
            )

        # Eqn 2.8 from [1]
        @self.Constraint(self.flowsheet().time)
        def flux_density_function_underflow(b, t):
            u = b.solid_fraction_underflow
            return b.flux_density_underflow[t] == Expr_if(
                IF=inequality(0, u[t], b.solid_fraction_max),
                THEN=b.v0[t] * u[t] * (1 - u[t] / b.solid_fraction_max) ** b.C
                + b.v1 * u[t] ** 2 * (b.solid_fraction_max - u[t]),
                ELSE=0 * units.m * units.s**-1,
            )

        # Modified from 2.23 and 2.33 from [1]
        @self.Constraint(self.flowsheet().time)
        def solids_continuity(b, t):
            return b.flow_vol_feed[t] * b.solid_fraction_feed[t] == (
                b.area * (b.flux_density_overflow[t] + b.flux_density_underflow[t])
                - b.flow_vol_overflow[t]
                * (b.solid_fraction_overflow[t] - b.solid_fraction_feed[t])
                + b.flow_vol_underflow[t]
                * (b.solid_fraction_underflow[t] - b.solid_fraction_feed[t])
            )

        @self.Constraint(self.flowsheet().time)
        def solids_conservation(b, t):
            return b.flow_vol_feed[t] * b.solid_fraction_feed[t] == (
                +b.flow_vol_overflow[t] * b.solid_fraction_overflow[t]
                + b.flow_vol_underflow[t] * b.solid_fraction_underflow[t]
            )

        @self.Constraint(self.flowsheet().time)
        def maximum_underflow_volume_fraction(b, t):
            return b.solid_fraction_underflow[t] <= b.solid_fraction_max

        @self.Constraint(self.flowsheet().time)
        def maximum_overflow_volume_fraction(b, t):
            return b.solid_fraction_overflow[t] <= b.solid_fraction_max

        @self.Constraint(self.flowsheet().time)
        def inlet_volume_fraction(b, t):
            return b.solid_inlet_state[t].flow_vol == (
                b.solid_fraction_feed[t]
                * (
                    b.solid_inlet_state[t].flow_vol
                    + units.convert(
                        b.liquid_inlet_state[t].flow_vol, to_units=uom.FLOW_VOL
                    )
                )
            )

        @self.Constraint(self.flowsheet().time)
        def underflow_volume_fraction(b, t):
            return b.solid_split.underflow_state[t].flow_vol == (
                b.solid_fraction_underflow[t]
                * (
                    b.solid_split.underflow_state[t].flow_vol
                    + units.convert(
                        b.liquid_split.underflow_state[t].flow_vol,
                        to_units=uom.FLOW_VOL,
                    )
                )
            )

        @self.Constraint(self.flowsheet().time)
        def stokes_law(b, t):
            # Assuming constant properties, source from feed states
            return 18 * b.v0[t] * units.convert(
                b.liquid_inlet_state[t].visc_d, to_units=uom.DYNAMIC_VISCOSITY
            ) == (
                (
                    b.solid_inlet_state[t].dens_mass
                    - units.convert(
                        b.liquid_inlet_state[t].dens_mass, to_units=uom.DENSITY_MASS
                    )
                )
                * units.convert(CONST.acceleration_gravity, to_units=uom.ACCELERATION)
                * b.particle_size[t] ** 2
            )

    def _get_performance_contents(self, time_point=0):
        return {
            "vars": {
                "Area": self.area,
                "Liquid Recovery": self.liquid_split.split_fraction[
                    time_point, "overflow"
                ],
                "Feed Solid Fraction": self.solid_fraction_feed[time_point],
                "Underflow Solid Fraction": self.solid_fraction_underflow[time_point],
                "particle size": self.particle_size[time_point],
                "v0": self.v0[time_point],
                "v1": self.v1,
                "C": self.C,
                "solid_fraction_max": self.solid_fraction_max,
            },
        }

    def _get_stream_table_contents(self, time_point=0):
        stream_attributes = {}
        stream_attributes["Units"] = {}

        sblocks = {
            "Feed Solid": self.solid_inlet_state,
            "Feed Liquid": self.liquid_inlet_state,
            "Underflow Solid": self.solid_split.underflow_state,
            "Underflow Liquid": self.liquid_split.underflow_state,
            "Overflow Solid": self.solid_split.overflow_state,
            "Overflow Liquid": self.liquid_split.overflow_state,
        }

        for n, v in sblocks.items():
            dvars = v[time_point].define_display_vars()

            stream_attributes[n] = {}

            for k in dvars:
                for i in dvars[k].keys():
                    stream_key = k if i is None else f"{k} {i}"

                    quant = report_quantity(dvars[k][i])

                    stream_attributes[n][stream_key] = quant.m
                    stream_attributes["Units"][stream_key] = quant.u

        return DataFrame.from_dict(stream_attributes, orient="columns")

    def initialize(self, **kwargs):
        raise NotImplementedError(
            "The Thickener0D unit model does not support the old initialization API. "
            "Please use the new API (InitializerObjects) instead."
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

__author__ = "Oluwamayowa Amusat"

# Imports from the python standard library
import os.path
import pickle

# Imports from third parties
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from scipy.optimize import basinhopping
import scipy.optimize as opt

from pyomo.core import Param, exp

# Imports from IDAES namespace
from idaes.core.surrogate.pysmo.sampling import FeatureScaling as fs


class MyBounds(object):
    """
    The Class MyBounds tests whether the reguularization parameter value is within the expected range.
     The class is initialized with the preset values in __init__; the __call__ function returns Booleans indicating whether the regularization parameter value is acceptable.
     The results of the __call__ function is fed into the Basinhopping algorithm using the accept_test parameter.
    """

    def __init__(self, xmax=None, xmin=None):
        if xmax is None:
            xmax = [1]
        if xmin is None:
            xmin = [1e-6]

        self.xmax = np.array(xmax)
        self.xmin = np.array(xmin)

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmax = bool(x[-1] <= self.xmax)
        tmin = bool(x[-1] >= self.xmin)
        return tmax and tmin


class KrigingModel:
    """
    The KrigingModel class trains a Kriging model for a training data set.

    The class must first be initialized by calling **KrigingModel**. Model training is then carried out by calling the ``training`` method.

    **KrigingModel** is able to generate either an interpolating or a regressing Kriging model depending on the settings used during initialization..

    **Example:**

    .. code-block:: python

        # Initialize the class
        >>> d = KrigingModel(training_data, numerical_gradients=True, regularization=True))
        >>> p = d.get_feature_vector()

        # Train Kriging model and predict output for an test data x_test
        >>> d.training()
        >>> predictions = d.predict_output(x_test)

    Args:
        XY_data (NumPy Array or Pandas Dataframe)   : The dataset for Kriging training. **XY_data** is expected to contain both the features (X) and output (Y) information, with the output values (Y) in the last column.

    Further details about the optional inputs may be found under the ``__init__`` method.

    """

    def __init__(
        self,
        XY_data,
        numerical_gradients=True,
        regularization=True,
        fname=None,
        overwrite=False,
    ):
        """
        Initialization of **KrigingModel** class.

        Args:
            XY_data (NumPy Array or Pandas Dataframe)   : The dataset for Kriging training. **XY_data** is expected to contain feature and output data, with the output values (y) in the last column.

        Keyword Args:
            numerical_gradients(bool)               : Whether or not numerical gradients should be used in training. This choice determines the algorithm used to solve the problem.

                                                            - numerical_gradients = True: The problem is solved with BFGS using central differencing with a step size of :math:`10^{-6}` to evaluate numerical gradients.
                                                            - numerical_gradients = False: The problem is solved with Basinhopping, a stochastic optimization algorithm.

            regularization(bool)                    :  This option determines whether or not regularization is considered during Kriging training. Default is True.

                                                            - When regularization is turned off, the model generates an interpolating kriging model.

        Returns:
            self object with the input information and settings.

        Raises:
            ValueError: - The input dataset is of the wrong type (not a NumPy array or Pandas Dataframe)

            Exception:  - numerical_gradients is not boolean

            Exception:  - regularization is not boolean

        **Example:**

        .. code-block:: python

                # Initialize Kriging class with no numerical gradients - solution algorithm will be Basinhopping
                >>> d = KrigingModel(XY_data, numerical_gradients=False)

        """
        if not isinstance(overwrite, bool):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("overwrite must be boolean.")
        self.overwrite = overwrite
        if fname is None:
            fname = "solution.pickle"
            self.filename = "solution.pickle"
        elif (
            not isinstance(fname, str)
            or os.path.splitext(fname)[-1].lower() != ".pickle"
        ):
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                'fname must be a string with extension ".pickle". Please correct.'
            )
        if (
            os.path.exists(fname) and overwrite is True
        ):  # Explicit overwrite done by user
            print(
                "Warning:",
                fname,
                "already exists; previous file will be overwritten.\n",
            )
            self.filename = fname
        elif os.path.exists(fname) and overwrite is False:  # User is not overwriting
            self.filename = (
                os.path.splitext(fname)[0]
                + "_v_"
                + pd.Timestamp.today().strftime("%m-%d-%y_%H%M%S")
                + ".pickle"
            )
            print(
                "Warning:",
                fname,
                'already exists; results will be saved to "',
                self.filename,
                '".\n',
            )
            # self.filename = 'solution.pickle'
        elif os.path.exists(fname) is False:
            self.filename = fname

        # Check data types and shapes
        if isinstance(XY_data, pd.DataFrame):
            xy_data = XY_data.values
            self.x_data_columns = list(XY_data.columns)[:-1]
        elif isinstance(XY_data, np.ndarray):
            xy_data = XY_data
            self.x_data_columns = list(range(XY_data.shape[1] - 1))
        else:
            raise ValueError('Pandas dataframe or numpy array required for "XY_data".')

        self.x_data = xy_data[:, :-1].reshape(xy_data.shape[0], xy_data.shape[1] - 1)
        self.y_data = xy_data[:, -1].reshape(xy_data.shape[0], 1)
        self.num_vars = self.x_data.shape[1] + 1  # thetas and reg parameter only
        x_data_scaled, self.x_data_min, self.x_data_max = fs.data_scaling_minmax(
            self.x_data
        )
        self.x_data_scaled = x_data_scaled.reshape(self.x_data.shape)

        if isinstance(numerical_gradients, bool):
            self.num_grads = numerical_gradients
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("numerical_gradients must be boolean.")

        if isinstance(regularization, bool):
            self.regularization = regularization
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("Choice of regularization must be boolean.")

        # Results
        self.optimal_weights = None
        self.optimal_p = None
        self.optimal_mean = None
        self.optimal_variance = None
        self.regularization_parameter = None
        self.optimal_covariance_matrix = None
        self.covariance_matrix_inverse = None
        self.optimal_y_mu = None
        self.output_predictions = None
        self.training_R2 = None
        self.training_rmse = None

    @staticmethod
    def covariance_matrix_generator(x, theta, reg_param, p):
        """
        The covariance_matrix_generator method generates the regularized co-variance matrix for a Kriging model

        Args:
            x                       : scaled features data
            theta                   : Kriging weights
            reg_param               : regularization parameter
            p                       : Kriging exponent, fixed at 2 for smoothness.

        Returns:
            cov_matrix              : Regularized co-variance matrix

        """
        distance_matrix = np.zeros((x.shape[0], x.shape[0]))
        for i in range(0, x.shape[0]):
            distance_matrix[i, :] = (
                np.matmul(((np.abs(x[i, :] - x)) ** p), theta)
            ).transpose()
        cov_matrix = np.exp(-1 * distance_matrix)
        cov_matrix = cov_matrix + reg_param * np.eye(
            cov_matrix.shape[0]
        )  # Regularization parameter addition, see Forrester book
        return cov_matrix

    @staticmethod
    def covariance_inverse_generator(x):
        """
        The covariance_inverse_generator method generates the inverse of the regularized co-variance matrix for a Kriging model

        Args:
            x                       : Regularized co-variance matrix

        Returns:
            inverse_x               : Inverse of regularized co-variance matrix

        """
        try:
            inverse_x = np.linalg.inv(x)
        except np.linalg.LinAlgError:
            inverse_x = np.linalg.pinv(x)
        return inverse_x

    @staticmethod
    def kriging_mean(cov_inv, y):
        """
        The kriging_mean method calculates the MLE estimate of the mean.

        Args:
            cov_inv (NumPy Array)           : Inverse of the co-variance matrix
            y (NumPy Array)                 : Output values of the training data


        Returns:
            kriging_mean                    : MLE estimate of the Kriging mean

        Reference:
            [1] Forrester et al.'s book "Engineering Design via Surrogate Modelling: A Practical Guide",
                https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470770801

        """
        ones_vec = np.ones((y.shape[0], 1))
        kriging_mean = np.matmul(
            np.matmul(ones_vec.transpose(), cov_inv), y
        ) / np.matmul(np.matmul(ones_vec.transpose(), cov_inv), ones_vec)
        # kriging_mean = np.matmul(ones_vec.transpose(), np.matmul(cov_inv, y)) / np.matmul(ones_vec.transpose(), np.matmul(cov_inv, ones_vec))
        return kriging_mean

    @staticmethod
    def y_mu_calculation(y, mu):
        """
        The y_mu_calculation method calculates the deviation of each output value from the MLE estimate of the mean, mu

        """
        y_mu = y - mu * np.ones((y.shape[0], 1))
        return y_mu

    @staticmethod
    def kriging_sd(cov_inv, y_mu, ns):
        """
        The kriging_sd method calculates the MLE estimate of the Kriging variance.

        Args:
            cov_inv (NumPy Array)           : Inverse of the co-variance matrix
            y_mu (NumPy Array)              : Deviation of y from the Kriging mean estimate (y-mean)


        Returns:
            kriging_sd                      : MLE estimate of the Kriging variance

        Reference:
            [1] Forrester et al.'s book "Engineering Design via Surrogate Modelling: A Practical Guide",
                https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470770801

        """
        sigma_sq = np.matmul(np.matmul(y_mu.transpose(), cov_inv), y_mu) / ns
        # sigma_sq = np.matmul(y_mu.transpose(), np.matmul(cov_inv, y_mu)) / ns
        return sigma_sq

    @staticmethod
    def print_fun(x, f, accepted):
        print("at minimum %.4f accepted %d" % (f, int(accepted)))

    def objective_function(self, var_vector, x, y, p):
        """
        The objective_function method calculates the concentrated likelihood function

        Args:
            var_vector(NumPy Array)        : Numpy array containing the Kriging parameters (Kriging weights and regularization parameter)
            x(NumPy Array)                 : Scaled version of input features/variables
            y(NumPy Array)                 : Output variable y (unscaled)
            p(float)                      : Kriging model exponent (fixed to 2) to ensure model smoothness

        Returns:
            conc_log_like(float)          : Concentrated likelihood value. Function incurs a large penalty (10000) when co-variance matrix is non-positive definite

        Reference:
            [1] Forrester et al.'s book "Engineering Design via Surrogate Modelling: A Practical Guide",
                https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470770801

        """
        theta = var_vector[:-1]
        reg_param = var_vector[-1]
        theta = 10**theta  # Assumes log(theta) provided
        ns = y.shape[0]
        cov_mat = self.covariance_matrix_generator(x, theta, reg_param, p)
        try:  # Check Cholesky factorization
            L = np.linalg.cholesky(cov_mat)
            lndetcov = 2 * np.sum(
                np.log(np.abs(np.diag(L)))
            )  # Approximation to 2nd term from Forrester book, making use of the Ch. factorization
            cov_inv = self.covariance_inverse_generator(cov_mat)
            km = self.kriging_mean(cov_inv, y)
            y_mu = self.y_mu_calculation(y, km)
            ssd = self.kriging_sd(cov_inv, y_mu, ns)
            # log_like = (0.5 * ns * np.log(ssd)) + (0.5 * np.log(np.abs(np.linalg.det(cov_mat))))
            log_like = (0.5 * ns * np.log(ssd)) + (0.5 * lndetcov)
            conc_log_like = log_like[0, 0]
        except Exception:  # pylint: disable=W0703
            # When Cholesky fails - non-positive definite covariance matrix
            conc_log_like = 1e4
        return conc_log_like

    def numerical_gradient(self, var_vector, x, y, p):
        """
        The numerical_gradient method calculates numerical gradients for the Kriging hyperparameters
        via central differencing,

        grad(theta) = (f(theta + eps) - f(theta - eps))/(2 * eps)

        Args:
            var_vector(NumPy Array): Numpy array containing the Kriging parameters (Kriging weights and
                regularization parameter)
            x(NumPy Array): Scaled version of input features/variables
            y(NumPy Array): Output variable y (unscaled)
            p(float): Kriging model exponent (fixed to 2) to ensure model smoothness

        Returns:
            grad_vec(NumPy Array): Array of the gradients of the variables in var_vector

        """
        eps = 1e-6
        grad_vec = np.zeros(
            len(var_vector),
        )
        for i in range(0, len(var_vector)):  # pylint: disable=consider-using-enumerate
            var_vector_plus = np.copy(var_vector)
            var_vector_plus[i] = var_vector[i] + eps
            var_vector_minus = np.copy(var_vector)
            var_vector_minus[i] = var_vector[i] - eps
            # of = self.objective_function(var_vector, x, y, p)
            of_plus = self.objective_function(var_vector_plus, x, y, p)
            of_minus = self.objective_function(var_vector_minus, x, y, p)
            grad_current = (of_plus - of_minus) / (2 * eps)
            grad_vec[i,] = grad_current
        if self.regularization is False:
            grad_vec[-1,] = 0
        return grad_vec

    def parameter_optimization(self, p):
        """
        Parameter (theta) optimization using BFGS or Basinhopping algorithm. This is the core of the Kriging Class.
        Algorithm used will depend on whether the numerical_gradients was set to True or False.
        """
        initial_value_list = np.random.randn(
            self.num_vars - 1,
        )
        initial_value_list = initial_value_list.tolist()
        initial_value_list.append(1e-4)
        initial_value = np.array(initial_value_list)
        initial_value = initial_value.reshape(
            initial_value.shape[0],
        )
        # Create bounds for variables. All logthetas btw (-4, 4), reg param between (1e-9, 0.1)
        bounds = []
        for i in range(0, len(initial_value_list)):
            if i == len(initial_value_list) - 1:
                if self.regularization is True:
                    bounds.append((1e-6, 0.1))
                else:
                    bounds.append((1e-10, 1e-10))
            else:
                bounds.append((-3, 3))
        bounds = tuple(bounds)

        if self.num_grads:
            print("Optimizing kriging parameters using L-BFGS-B algorithm...")
            other_args = (self.x_data_scaled, self.y_data, p)
            # opt_results = opt.minimize(self.objective_function, initial_value, args=other_args, method='L-BFGS-B', jac=self.numerical_gradient, bounds=bounds, options={'gtol': 1e-7}) #, 'disp': True})
            opt_results1 = opt.minimize(
                self.objective_function,
                initial_value,
                args=other_args,
                method="tnc",
                jac=self.numerical_gradient,
                bounds=bounds,
                options={"gtol": 1e-7},
            )
            opt_results2 = opt.minimize(
                self.objective_function,
                initial_value,
                args=other_args,
                method="L-BFGS-B",
                jac=self.numerical_gradient,
                bounds=bounds,
                options={"gtol": 1e-7},
            )  # , 'disp': True})
            if opt_results1.fun < opt_results2.fun:
                opt_results = opt_results1
            else:
                opt_results = opt_results2
        else:
            print("Optimizing Kriging parameters using Basinhopping algorithm...")
            other_args = {
                "args": (self.x_data_scaled, self.y_data, p),
                "bounds": bounds,
            }
            # other_args = {"args": (self.x_data, self.y_data, p)}
            mybounds = MyBounds()  # Bounds on regularization parameter
            opt_results = basinhopping(
                self.objective_function,
                initial_value_list,
                minimizer_kwargs=other_args,
                niter=250,
                disp=True,
                accept_test=mybounds,
            )  # , interval=5)
        return opt_results

    def optimal_parameter_evaluation(self, var_vector, p):
        """
        The optimal_parameter_evaluation method  evaluates the values of all the parameters of the final Kriging model.
        For an input set of Kriging parameters var_vector and p, it:

            (1) Generates the covariance matrix by calling covariance_matrix_generator
            (2) Finds the co-variance matrix inverse
            (3) Evaluates the Kriging mean and variance
            (4) Evaluates the deviation of each training point from the Kriging mean

        Args:
            var_vector              : Optimal Kriging parameters (weights + regularization parameter)
            p                       : Krigng exponents

        Returns:
            theta                   : Optimal Kriging weights for each variable
            reg_param               : Optimal regularization parameter
            mean                    : Final MLE estimate of the mean
            variance                : Final MLE estimate of the variance
            cov_mat                 : Co-variance matrix of the final model
            cov_inv                 : Inverse of final co-variance matrix
            y_mu                    : Deviation of each output value in the training data from the Kriging mean.

        """
        theta = var_vector[:-1]
        reg_param = var_vector[-1]
        theta = (
            10**theta
        )  # Assumes log(theta) provided. Ensures that theta is always positive
        ns = self.y_data.shape[0]
        cov_mat = self.covariance_matrix_generator(
            self.x_data_scaled, theta, reg_param, p
        )
        cov_inv = self.covariance_inverse_generator(cov_mat)
        mean = self.kriging_mean(cov_inv, self.y_data)
        y_mu = self.y_mu_calculation(self.y_data, mean)
        variance = self.kriging_sd(cov_inv, y_mu, ns)
        print(
            "\nFinal results\n================\nTheta:",
            theta,
            "\nMean:",
            mean,
            "\nRegularization parameter:",
            reg_param,
        )
        return theta, reg_param, mean, variance, cov_mat, cov_inv, y_mu

    @staticmethod
    def error_calculation(theta, p, mean, cov_inv, y_mu, x, y_data):
        """
        This method calculates the SSE and RMSE errors between the actual and predicted output values,
             ss_error = sum of squared errors / number of samples
             rmse_error = sqrt(sum of squared errors / number of samples)

        Args:
            theta           : Kriging weights
            p               : Kriging exponents
            mean            : MLE estimate of the Kriging mean
            cov_inv         : Inverse of the co-variance matrix of the current solution
            y_mu            : Deviation of y valuers from the mean estimate
            x               : Input test data
            y_data          : Actual outputs corresponding to input test data x

        Returns:
            ss_error        : The average sum of squared errors
            rmse_error      : The root-mean-squared error (RMSE)
            y_prediction    : Predicted values of y

        """
        y_prediction = np.zeros((x.shape[0], 1))
        for i in range(0, x.shape[0]):
            cmt = (np.matmul(((np.abs(x[i, :] - x)) ** p), theta)).transpose()
            cov_matrix_tests = np.exp(-1 * cmt)
            y_pred_val = mean + np.matmul(
                np.matmul(cov_matrix_tests.transpose(), cov_inv), y_mu
            )
            y_prediction[i, 0] = y_pred_val.item()
        ss_error = (1 / y_data.shape[0]) * (np.sum((y_data - y_prediction) ** 2))
        rmse_error = np.sqrt(ss_error)
        return ss_error, rmse_error, y_prediction

    @staticmethod
    def r2_calculation(y_true, y_predicted):
        """
        ``r2_calculation`` returns the :math:`R^{2}` as a measure-of-fit between the true and predicted values of the output variable.

        Args:
            y_true(NumPy Array)             : Vector of actual values of the output variable
            y_predicted(NumPy Array)        : Vector of predictions for the output variable based on the surrogate

        Returns:
            float                           : :math:`R^{2}` measure-of-fit between actual and predicted data

        """
        y_true = y_true.reshape(y_true.shape[0], 1)
        y_predicted = y_predicted.reshape(y_predicted.shape[0], 1)
        input_y_mean = np.mean(y_true, axis=0)
        ss_total = np.sum((y_true - input_y_mean) ** 2)
        ss_residual = np.sum((y_predicted - y_true) ** 2)
        r_square = 1 - (ss_residual / ss_total)
        return r_square

    def predict_output(self, x_pred):
        """
        The ``predict_output`` method generates output predictions for input data x_pred based a previously trained Kriging model.

        Args:
            x_pred(NumPy Array)             : Array of designs for which the output is to be evaluated/predicted.

        Returns:
             NumPy Array                    : Output variable predictions based on the Kriging model.

        """
        x_pred_scaled = (x_pred - self.x_data_min) / (self.x_data_max - self.x_data_min)
        x_pred = x_pred_scaled.reshape(x_pred.shape)
        if x_pred.ndim == 1:
            x_pred = x_pred.reshape(1, len(x_pred))
        y_pred = np.zeros((x_pred.shape[0], 1))
        for i in range(0, x_pred.shape[0]):
            cmt = (
                np.matmul(
                    ((np.abs(x_pred[i, :] - self.x_data_scaled)) ** self.optimal_p),
                    self.optimal_weights,
                )
            ).transpose()
            cov_matrix_tests = np.exp(-1 * cmt)
            y_pred_val = self.optimal_mean + np.matmul(
                np.matmul(cov_matrix_tests.transpose(), self.covariance_matrix_inverse),
                self.optimal_y_mu,
            )
            y_pred[i, 0] = y_pred_val.item()
        return y_pred

    def training(self):
        r"""
        Main function for Kriging training.

        To train the Kriging model:
            (1) The Kriging exponent :math:`\tau_{i}` is fixed at 2.
            (2) The optimal Kriging hyperparameters :math:`\left(\mu,\sigma^{2},\theta_{1},\ldots,\theta_{n}\right)` are evaluated by calling the ``optimal_parameter_evaluation`` method using either BFGS or Basinhopping.
            (3) The training predictions, prediction errors and r-square coefficient of fit are evaluated by calling the functions ``error_calculation`` and ``self.r2_calculation``
            (4) A results object is generated by calling the ``ResultsReport`` class.

        Returns:
            tuple   : self object (**results**) containing the all information about the best Kriging model obtained, including:
                - the Kriging model hyperparameters  (**results.optimal_weights**),
                - when relevant, the optimal regularization parameter found :math:`\lambda` (**results.regularization_parameter**),
                - the Kriging mean (**results.optimal_mean**),
                - the Kriging variance (**results.optimal_variance**),
                - the Kriging model regularized co-variance matrix (**results.optimal_covariance_matrix**),
                - the inverse of the co-variance matrix (**results.covariance_matrix_inverse**),
                - the RBF predictions for the training data (**results.output_predictions**),
                - the RMSE of the training output predictions (**results.training_rmse**), and
                - the :math:`R^{2}` value on the training data (**results.R2**)

        """

        # Create p values, for now fixed at p=2. Arraying p makes the code significantly (at least 7x slower)
        p = 2
        # Solve optimization problem
        bh_results = self.parameter_optimization(p)
        # Calculate other variables and parameters
        (
            optimal_theta,
            optimal_reg_param,
            optimal_mean,
            optimal_variance,
            optimal_cov_mat,
            opt_cov_inv,
            optimal_ymu,
        ) = self.optimal_parameter_evaluation(bh_results.x, p)
        # Training performance
        (
            training_ss_error,  # pylint: disable=unused-variable
            rmse_error,
            y_training_predictions,
        ) = self.error_calculation(
            optimal_theta,
            p,
            optimal_mean,
            opt_cov_inv,
            optimal_ymu,
            self.x_data_scaled,
            self.y_data,
        )
        r2_training = self.r2_calculation(self.y_data, y_training_predictions)

        # Return results
        self.optimal_weights = optimal_theta
        self.optimal_p = p
        self.optimal_mean = optimal_mean
        self.optimal_variance = optimal_variance
        self.regularization_parameter = optimal_reg_param
        self.optimal_covariance_matrix = optimal_cov_mat
        self.covariance_matrix_inverse = opt_cov_inv
        self.optimal_y_mu = optimal_ymu
        self.output_predictions = y_training_predictions
        self.training_R2 = r2_training
        self.training_rmse = rmse_error
        self.pickle_save({"model": self})
        return self

    def generate_expression(self, variable_list):
        """
        The ``generate_expression`` method returns the Pyomo expression for the Kriging model trained.

        The expression is constructed based on the supplied list of variables **variable_list** and the results of the previous Kriging training process.

        Args:
            variable_list(list)           : List of input variables to be used in generating expression. This can be the a list generated from the output of ``get_feature_vector``.  The user can also choose to supply a new list of the appropriate length.

        Returns:
            Pyomo Expression              : Pyomo expression of the Kriging model based on the variables provided in **variable_list**

        """
        t1 = np.array([variable_list], dtype="object")
        # Reshaping of variable array is necessary when input variables are Pyomo scalar variables
        t1 = t1.reshape(1, len(variable_list)) if t1.ndim > 2 else t1

        phi_var = []
        for i in range(0, self.x_data.shape[0]):
            curr_term = sum(
                self.optimal_weights[j]
                * (
                    (
                        (t1[0, j] - self.x_data_min[0, j])
                        / (self.x_data_max[0, j] - self.x_data_min[0, j])
                    )
                    - self.x_data_scaled[i, j]
                )
                ** self.optimal_p
                for j in range(0, self.x_data.shape[1])
            )

            curr_term = exp(-curr_term)
            phi_var.append(curr_term)
        phi_var_array = np.asarray(phi_var)

        phi_inv_times_y_mu = np.matmul(
            self.covariance_matrix_inverse, self.optimal_y_mu
        )
        phi_inv_times_y_mu = phi_inv_times_y_mu.reshape(
            phi_inv_times_y_mu.shape[0],
        )
        kriging_expr = self.optimal_mean[0, 0]
        kriging_expr += sum(
            w * t
            for w, t in zip(
                np.nditer(phi_inv_times_y_mu),
                np.nditer(phi_var_array, flags=["refs_ok"]),
            )
        )
        return kriging_expr

    def get_feature_vector(self):
        """

        The ``get_feature_vector`` method generates the list of regression features from the column headers of the input dataset.

        Returns:
            Pyomo IndexedParam  : An indexed parameter list of the variables supplied in the original data

        """
        p = Param(self.x_data_columns, mutable=True, initialize=0)
        p.index_set().construct()
        p.construct()
        self.feature_list = p
        return p

    def pickle_save(self, solutions):
        """
        The poly_training method saves the results of the run in a pickle object. It saves an object with two elements: the setup (index[0]) and the results (index[1]).
        """
        try:
            filehandler = open(self.filename, "wb")
            pickle.dump(solutions, filehandler)
            print("\nResults saved in ", str(self.filename))
        except:
            raise IOError("File could not be saved.")

    @staticmethod
    def pickle_load(solution_file):
        """
        pickle_load loads the results of a saved run 'file.obj'.])

        Input arguments:
                solution_file            : Pickle object file containing previous solution to be loaded.

        """
        try:
            filehandler = open(solution_file, "rb")
            return pickle.load(filehandler)
        except:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("File could not be loaded.")

    def parity_residual_plots(self):
        """

        inputs:

        Returns:

        """

        fig1 = plt.figure(figsize=(16, 9), tight_layout=True)
        ax = fig1.add_subplot(121)
        ax.plot(self.y_data, self.y_data, "-")
        ax.plot(self.y_data, self.output_predictions, "o")
        ax.set_xlabel(r"True data", fontsize=12)
        ax.set_ylabel(r"Surrogate values", fontsize=12)
        ax.set_title(r"Parity plot", fontsize=12)

        ax2 = fig1.add_subplot(122)
        ax2.plot(
            self.y_data,
            self.y_data - self.output_predictions,
            "s",
            mfc="w",
            mec="m",
            ms=6,
        )
        ax2.axhline(y=0, xmin=0, xmax=1)
        ax2.set_xlabel(r"True data", fontsize=12)
        ax2.set_ylabel(r"Residuals", fontsize=12)
        ax2.set_title(r"Residual plot", fontsize=12)

        plt.show()

        return

    def _report(self):
        ## Will only work with Python > 3.5
        variable_headers = self.get_feature_vector()
        var_list = []
        for i in variable_headers:
            var_list.append(variable_headers[i])
        eqn = self.generate_expression(var_list)

        double_line = "=" * 120
        s = (
            f"\n{double_line}"
            f"\nResults of Kriging run:\n"
            f"\nKriging mean                     : {self.optimal_mean}\n"
            f"Kriging variance                 : {self.optimal_variance}\n"
            f"Kriging weights                  : {self.optimal_weights}\n"
            f"Regularization parameter         : {self.regularization_parameter}\n"
            f"Number of terms in Kriging model : {self.optimal_y_mu.size + 1}\n"
            f"\nKriging Expression:\n"
            f"--------------------\n"
            f"\n{eqn}\n"
            f"--------------------------\n"
            f"\nModel training errors:"
            f"\n-----------------------\n"
            f"Mean Squared Error (MSE)         : {self.training_rmse ** 2}\n"
            f"Root Mean Squared Error (RMSE)   : {self.training_rmse}\n"
            f"Goodness of fit (R2)             : {self.training_R2}\n"
            f"\n{double_line}"
        )
        return s

    def print_report(self):
        s = self._report()
        print(s)

    def _repr_pretty_(self, p, cycle=False):
        s = self._report()
        p.text(s)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Natural gas property package with a single (pseudo) component for use
in isothermal unit models.

Data sources:
    [1] Stochastic Optimal Control Model for Natural Gas Network
        Operations. V. Zavala, 2014, Comp. Chem. Eng.

"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

from pyomo.core.base.units_container import units as pyunits
from pyomo.core.base.var import Var
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.expression import Expression
from pyomo.core.base.param import Param
from pyomo.core.expr import sqrt

# Import IDAES cores
from idaes.core import (
    PhysicalParameterBlock,
)
import idaes.logger as idaeslog

from idaes.core import (
    declare_process_block_class,
    VaporPhase,
    Component,
    MaterialFlowBasis,
    StateBlock,
    StateBlockData,
)
from idaes.core.util.constants import Constants

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("NaturalGasParameterBlock")
class NaturalGasParameterBlockData(PhysicalParameterBlock):
    """
    Property package for natural gas with a single pseudo-component.
    """

    def build(self):
        super(NaturalGasParameterBlockData, self).build()
        self._state_block_class = NaturalGasStateBlock
        self.Vap = VaporPhase()

        self.dens_nominal = Param(
            initialize=0.72,
            units=pyunits.kg / pyunits.m**3,
            doc="Density of the gas a standard temperature and pressure",
            # Used to convert mass flow rate between kg/hr and SCM/hr
        )

        nat_gas_data = {"mw": (18.0, pyunits.kg / pyunits.kmol)}
        nat_gas_config = {"parameter_data": nat_gas_data}
        self.natural_gas = Component(**nat_gas_config)
        ng = self.natural_gas

        # Set up dictionaries we will rely on to compute heat capacities.
        kJkmolK = pyunits.kJ / pyunits.kmol / pyunits.K
        kJkmolK2 = pyunits.kJ / pyunits.kmol / pyunits.K**2
        kJkmolK3 = pyunits.kJ / pyunits.kmol / pyunits.K**3
        kJkmolK4 = pyunits.kJ / pyunits.kmol / pyunits.K**4
        ng.cp_mol_coef_A = Param(initialize=2.34 * 18.0, units=kJkmolK)
        ng.cp_mol_coef_B = Param(initialize=0.0, units=kJkmolK2)
        ng.cp_mol_coef_C = Param(initialize=0.0, units=kJkmolK3)
        ng.cp_mol_coef_D = Param(initialize=0.0, units=kJkmolK4)

        ng.cv_mol_coef_A = Param(initialize=1.85 * 18.0, units=kJkmolK)
        ng.cv_mol_coef_B = Param(initialize=0.0, units=kJkmolK2)
        ng.cv_mol_coef_C = Param(initialize=0.0, units=kJkmolK3)
        ng.cv_mol_coef_D = Param(initialize=0.0, units=kJkmolK4)

        self.temperature_ref = Param(initialize=298.15, units=pyunits.K)

    @classmethod
    def define_metadata(cls, obj):
        kghr = pyunits.kg / pyunits.hr
        nondim = pyunits.dimensionless
        kgkmolK = pyunits.kg / pyunits.kmol / pyunits.K
        kmolm3 = pyunits.kmol / pyunits.m**3
        kgm3 = pyunits.kg / pyunits.m**3
        kmolhr = pyunits.kmol / pyunits.hr
        obj.add_properties(
            {
                # What do the units in this dict get used for? What if they're
                # different than the units we define for each variable in the
                # StateBlockData methods?
                "flow_mol": {"method": None, "units": kmolhr},
                "pressure": {"method": None, "units": pyunits.bar},
                "temperature": {"method": None, "units": pyunits.K},
                "mole_frac_comp": {"method": None, "units": pyunits.dimensionless},
                "flow_mol_comp": {"method": "_flow_mol_comp", "units": kmolhr},
                "mw": {"method": "_mw", "units": pyunits.kg / pyunits.kmol},
                "flow_mass": {"method": "_flow_mass", "units": kghr},
                "cp_mol_comp": {"method": "_cp_mol_comp", "units": kgkmolK},
                "cp_mol": {"method": "_cp_mol", "units": kgkmolK},
                "cp_mass": {"method": "_cp_mass", "units": kgkmolK},
                "cv_mol_comp": {"method": "_cv_mol_comp", "units": kgkmolK},
                "cv_mol": {"method": "_cv_mol", "units": kgkmolK},
                "cv_mass": {"method": "_cv_mass", "units": kgkmolK},
                "heat_capacity_ratio": {
                    "method": "_heat_capacity_ratio",
                    "units": nondim,
                },
                "heat_capacity_ratio_phase": {
                    "method": "_heat_capacity_ratio_phase",
                    "units": nondim,
                },
                "compress_fact": {"method": "_compress_fact", "units": nondim},
                "dens_mol": {"method": "_dens_mol", "units": kmolm3},
                "dens_mol_comp": {"method": "_dens_mol_comp", "units": kmolm3},
                "dens_mass": {"method": "_dens_mass", "units": kgm3},
            }
        )
        # NOTE: We do not implement enthalpy as we are not yet using this
        # property package with a non-isothermal pipeline.

        obj.define_custom_properties(
            {
                "speed_of_sound": {
                    "method": "_speed_of_sound",
                    "units": obj.derived_units.VELOCITY,
                },
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.hr,
                "length": pyunits.km,
                "mass": pyunits.kg,
                "amount": pyunits.kmol,
                "temperature": pyunits.K,
                # I would like to specify some reasonable units for area,
                # energy, and pressure, but this is not supported in IDAES.
                # "area": pyunits.m**2,
                # "energy": pyunits.kJ,
                # "pressure": pyunits.bar,
            }
        )


class NaturalGasStateBlock(StateBlock):
    # This is confusing.
    # (a) Not sure why this class is necessary when I don't want
    #     to attach any methods.
    # (b) Now is NaturalGasStateBlock defined twice? Once here and once
    #     by NaturalGasStateBlock? This appears to be the case. This
    #     class gets overridden...
    # declare_process_block_class on its own is confusing...
    pass


@declare_process_block_class(
    "NaturalGasStateBlock",
    block_class=NaturalGasStateBlock,
)
class NaturalGasStateBlockData(StateBlockData):
    def build(self):
        super().build()

        # TODO: Initialize to 10 (1e6 SCM)/day
        self.flow_mol = Var(
            initialize=300.0,
            doc="Molar flow rate",
            units=pyunits.kmol / pyunits.hr,
        )
        self.pressure = Var(
            initialize=50.0,
            doc="Gas pressure",
            units=pyunits.bar,
        )
        self.temperature = Var(
            initialize=298.15,
            doc="Gas temperature",
            units=pyunits.K,
        )
        component_list = self.config.parameters.component_list
        self.mole_frac_comp = Var(
            component_list,
            initialize=1.0 / len(component_list),
            doc="Component mole fractions within the gas",
            units=pyunits.dimensionless,
        )

        if not self.config.defined_state:
            # For a single-phase property package, should we just always
            # "fix" mole fraction? Probably not, as it will violate assumptions
            # when we combine multiple units (ports) as long as mole_frac_comp
            # is a state variable, which I would rather not change.
            def sum_component_eq_rule(b):
                return 1.0 == sum(b.mole_frac_comp[j] for j in component_list)

            self.sum_component_eq = Constraint(
                rule=sum_component_eq_rule,
                doc=(
                    "Enforces that the sum of mole fractions equals one when\n"
                    "state variables are not already elsewhere defined."
                ),
            )

    def _flow_mol_comp(self):
        params = self.config.parameters
        component_list = params.component_list

        def flow_mol_comp_rule(b, j):
            return self.flow_mol * self.mole_frac_comp[j]

        self.flow_mol_comp = Expression(
            component_list,
            rule=flow_mol_comp_rule,
            doc="Molar flow rate of a particular component",
        )

    def _mw(self):
        params = self.config.parameters
        component_list = params.component_list
        mole_frac = self.mole_frac_comp
        mw_comp = {j: params.get_component(j).mw for j in component_list}
        self.mw = Expression(
            expr=sum(mole_frac[j] * mw_comp[j] for j in component_list),
            doc="Average molecular weight of the gas",
        )

    def _flow_mass(self):
        # I would like flow_mass to be a variable, so I can fix it.
        # However, I leave it as an expression here for consistency with
        # generic property packages. A flow_mass variable will be added
        # by pipeline unit models so that its derivatives may be computed.
        self.flow_mass = Expression(
            expr=self.mw * self.flow_mol,
            doc="Mass flow rate of the gas",
        )

    def _cp_mol_comp(self):
        params = self.config.parameters
        component_list = params.component_list
        comp = {j: params.get_component(j) for j in component_list}

        def cp_mol_comp_rule(b, j):
            return (
                comp[j].cp_mol_coef_A
                + comp[j].cp_mol_coef_B * b.temperature
                + comp[j].cp_mol_coef_C * b.temperature**2
                + comp[j].cp_mol_coef_D * b.temperature**3
            )

        self.cp_mol_comp = Expression(
            component_list,
            rule=cp_mol_comp_rule,
            doc=(
                "Pure component constant-pressure molar heat capacity "
                "of each component"
            ),
        )

    def _cp_mol(self):
        component_list = self.config.parameters.component_list
        self.cp_mol = Expression(
            expr=sum(
                self.mole_frac_comp[j] * self.cp_mol_comp[j] for j in component_list
            ),
            doc="Constant-pressure molar heat capacity of the gas mixture",
        )

    def _cp_mass(self):
        self.cp_mass = Expression(
            expr=self.cp_mol / self.mw,
            doc="Constant-pressure specific heat capacity of the gas mixture",
        )

    def _cv_mol_comp(self):
        params = self.config.parameters
        component_list = params.component_list
        comp = {j: params.get_component(j) for j in component_list}

        def cv_mol_comp_rule(b, j):
            return (
                comp[j].cv_mol_coef_A
                + comp[j].cv_mol_coef_B * b.temperature
                + comp[j].cv_mol_coef_C * b.temperature**2
                + comp[j].cv_mol_coef_D * b.temperature**3
            )

        self.cv_mol_comp = Expression(
            component_list,
            rule=cv_mol_comp_rule,
            doc=(
                "Pure component constant-volume molar heat capacity "
                "of each component"
            ),
        )

    def _cv_mol(self):
        component_list = self.config.parameters.component_list
        self.cv_mol = Expression(
            expr=sum(
                self.mole_frac_comp[j] * self.cv_mol_comp[j] for j in component_list
            ),
            doc="Constant-volume molar heat capacity of the gas mixture",
        )

    def _cv_mass(self):
        self.cv_mass = Expression(
            expr=self.cv_mol / self.mw,
            doc="Constant-volume specific heat capacity of the gas mixture",
        )

    def _heat_capacity_ratio(self):
        self.heat_capacity_ratio = Expression(
            expr=self.cp_mass / self.cv_mass,
            doc=(
                "Ratio of constant-volume to constant-pressure heat "
                "capacities of the gas mixture"
            ),
        )

    def _heat_capacity_ratio_phase(self):
        # Pipeline unit models require a heat_capacity_ratio_phase attribute
        # for compatibility with generic property packages.
        self.heat_capacity_ratio_phase = Expression(
            self.config.parameters.phase_list,
            expr=self.heat_capacity_ratio.expr,
            doc=(
                "Ratio of constant-volume to constant-pressure heat "
                "capacities of the gas mixture"
            ),
        )

    def _compress_fact(self):
        # compress_fact is a param because here it is constant.
        # It could be a variable/expression, however, in a more complicated
        # model.
        self.compress_fact = Param(
            initialize=0.80,
            doc="Compressibility factor of the gas",
        )

    def _dens_mol(self):
        # Make molar density a variable as it is important enough that
        # somebody might want to fix it.
        self.dens_mol = Var(
            initialize=1.0,
            units=pyunits.kmol / pyunits.m**3,
            doc="Molar density of the gas phase",
        )
        pressure = self.pressure
        gas_const = Constants.gas_constant
        compress_fact = self.compress_fact
        temperature = self.temperature
        dens_mol_expr = pyunits.convert(
            pressure / gas_const / compress_fact / temperature,
            pyunits.kmol / pyunits.m**3,
        )
        self.dens_mol_eq = Constraint(
            expr=self.dens_mol == dens_mol_expr,
            doc=(
                "Equation used to calculate molar density -- "
                "ideal gas equation with a\ncompressibility factor"
            ),
        )

    def _dens_mol_comp(self):
        component_list = self.config.parameters.component_list

        def dens_mol_comp_rule(b, j):
            return self.mole_frac_comp[j] * self.dens_mol

        self.dens_mol_comp = Expression(
            component_list,
            rule=dens_mol_comp_rule,
            doc="Molar density of a particular component in the gas",
        )

    def _speed_of_sound(self):
        # Use a constraint here to make balance equation expressions more
        # legible.
        self.speed_of_sound = Var(
            initialize=300.0,
            units=pyunits.m / pyunits.s,
            doc="Speed of sound in the gas",
        )
        gas_const = pyunits.convert(
            Constants.gas_constant,
            pyunits.kJ / pyunits.kmol / pyunits.K,
        )
        speed_of_sound_expr = pyunits.convert(
            sqrt(
                self.heat_capacity_ratio
                * self.compress_fact
                * gas_const
                * self.temperature
                / self.mw
            ),
            pyunits.m / pyunits.s,
        )
        self.speed_of_sound_eq = Constraint(
            expr=self.speed_of_sound == speed_of_sound_expr,
            doc="Equation to calculate speed of sound",
        )

    def define_state_vars(self):
        return {
            "flow_mol": self.flow_mol,
            "pressure": self.pressure,
            "temperature": self.temperature,
            "mole_frac_comp": self.mole_frac_comp,
        }

    def get_material_flow_terms(self, p, j):
        return self.flow_mol * self.mole_frac_comp[j]

    def get_material_density_terms(self, p, j):
        # Converting to kmol/km^3 is a workaround -- really I would like to
        # change the units of area and material_holdup in the control volume.
        return pyunits.convert(
            self.dens_mol_comp[j],
            pyunits.kmol / pyunits.km**3,
        )

    def get_material_flow_basis(self):
        return MaterialFlowBasis.molar

    def get_enthalpy_flow_terms(self, p):
        return (
            (self.temperature - self.config.parameters.temperature_ref)
            * self.cp_mol
            * self.flow_mol
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for defining equilibrium reactions
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

from pyomo.environ import Param, units as pyunits, value

from idaes.core.util.math import smooth_max
from idaes.core.util.exceptions import ConfigurationError

from idaes.models.properties.modular_properties.base.utility import (
    get_concentration_term,
)


# ----------------------------------------------------------------------------
class power_law_equil:
    """Methods for power-law based equilibrium forms."""

    @staticmethod
    def build_parameters(rblock, config):
        pass

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        e = None

        if hasattr(b.params, "_electrolyte") and b.params._electrolyte:
            pc_set = b.params.true_phase_component_set
        else:
            pc_set = b.phase_component_set

        # Get reaction orders and construct power law expression
        for p, j in pc_set:
            o = rblock.reaction_order[p, j]

            if e is None and o.value != 0:
                e = get_concentration_term(b, r_idx)[p, j] ** o
            elif e is not None and o.value != 0:
                e = e * get_concentration_term(b, r_idx)[p, j] ** o

        return b.k_eq[r_idx] == e

    @staticmethod
    def calculate_scaling_factors(b, sf_keq):
        return sf_keq


# ----------------------------------------------------------------------------
class log_power_law_equil:
    """Methods for log formulation of power-law based equilibrium forms."""

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        e = None

        if hasattr(b.params, "_electrolyte") and b.params._electrolyte:
            pc_set = b.params.true_phase_component_set
        else:
            pc_set = b.phase_component_set

        # Get reaction orders and construct power law expression
        for p, j in pc_set:
            o = rblock.reaction_order[p, j]

            if e is None and o.value != 0:
                c = get_concentration_term(b, r_idx, log=True)[p, j]
                e = o * c
            elif e is not None and o.value != 0:
                c = get_concentration_term(b, r_idx, log=True)[p, j]
                e = e + o * c

        return b.log_k_eq[r_idx] == e

    @staticmethod
    def calculate_scaling_factors(b, sf_keq):
        return sf_keq


# ----------------------------------------------------------------------------
class solubility_product:
    """
    Complementarity formulation for solid precipitation
    Thanks to Larry Biegler for the formulation

    Like any phase equilibrium, solid precipitation is complicated by
    conditional constraint. For precipitation, the solubility product,
    Ksp = f(C) only holds if solids are present in the system. That is, if the
    system is subsaturated and no solids are present, Ksp >= f(C) instead.

    Letting S represent solids in the system and Q = Ksp - f(C), we have the
    following:

        * If S >= 0, Q = 0
        * If S = 0, Q >= 0

    Thus, only one of S and Q can be greater than zero at any time.
    This can be written in the form of a complementarity constraint as:

        * Q - MAX(0, Q-S) == 0

    where S is assumed to be the sum of the flowrates any solids formed in the
    reaction. This allows for multiple solid products, and only applies the
    solubility product if all flowrates are non-zero.

    Note that the smooth maximum approximation requires the use of a smoothing
    factor eps, and that the value of this factor needs to be adjusted for the
    specific problem. as an initial guideline, it is suggested that this be at
    least one to two orders of magnitude smaller than the solubility product.
    """

    @staticmethod
    def build_parameters(rblock, config):
        rblock.eps = Param(
            mutable=True, initialize=1e-4, doc="Smoothing parameter for smooth maximum"
        )

        rblock.s_norm = Param(
            mutable=True,
            initialize=1e-4,
            doc="Normalizing factor for solid precipitation term",
        )
        if hasattr(rblock, "k_eq_ref"):
            rblock.s_norm.set_value(value(rblock.k_eq_ref))

        rblock.s_scale = Param(
            mutable=True,
            initialize=1,
            doc="Scaling factor for solid precipitation term w.r.t saturated status Q = Ksp - f(C)",
        )

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        e = None
        s = None

        if hasattr(b.params, "_electrolyte") and b.params._electrolyte:
            pc_set = b.params.true_phase_component_set
        else:
            pc_set = b.phase_component_set

        # Get reaction orders and construct power law expression
        for p, j in pc_set:
            p_obj = b.state_ref.params.get_phase(p)

            # First, build E
            o = rblock.reaction_order[p, j]

            if e is None and o.value != 0:
                e = get_concentration_term(b, r_idx)[p, j] ** o
            elif e is not None and o.value != 0:
                e = e * get_concentration_term(b, r_idx)[p, j] ** o

            if p_obj.is_solid_phase():
                # If solid phase, identify S
                r_config = b.params.config.equilibrium_reactions[r_idx]

                try:
                    stoic = r_config.stoichiometry[p, j]
                    if s is None and stoic != 0:
                        s = b.state_ref.flow_mol_phase_comp[p, j]
                    elif s is not None and stoic != 0:
                        s += b.state_ref.flow_mol_phase_comp[p, j]

                except KeyError:
                    pass

        if s is None:
            # Catch for not finding a solid phase
            raise ConfigurationError(
                "{} did not find a solid phase component for precipitation "
                "reaction {}. This is likely due to the reaction "
                "configuration.".format(b.name, r_idx)
            )
        else:
            # Need to remove units as complementarity is not consistent
            sunits = pyunits.get_units(s)

            if sunits is not None:
                s = s / sunits

        s = rblock.s_scale * s / (s + rblock.s_norm)

        Q = b.k_eq[r_idx] - e

        # Need to remove units again
        Qunits = pyunits.get_units(b.k_eq[r_idx])

        if Qunits is not None:
            Q = Q / Qunits

        return Q - smooth_max(0, Q - s, rblock.eps) == 0

    @staticmethod
    def calculate_scaling_factors(b, sf_keq):
        return sf_keq


class log_solubility_product:
    """
    Complementarity formulation for solid precipitation
    Thanks to Larry Biegler for the formulation

    Like any phase equilibrium, solid precipitation is complicated by
    conditional constraint. For precipitation, the solubility product,
    Ksp = f(C) only holds if solids are present in the system. That is, if the
    system is subsaturated and no solids are present, Ksp >= f(C) instead. This
    class uses a log form to represent to equilibrium constraint.

    Letting S represent solids in the system and Q = ln(Ksp) - ln(f(C)), we
    have the following:

        * If S >= 0, Q = 0
        * If S = 0, Q >= 0

    Thus, only one of S and Q can be greater than zero at any time.
    This can be written in the form of a complementarity constraint as:

        * Q - MAX(0, Q-S) == 0

    where S is assumed to be the sum of the flowrates any solids formed in the
    reaction. This allows for multiple solid products, and only applies the
    solubility product if all flowrates are non-zero.

    Note that the smooth maximum approximation requires the use of a smoothing
    factor eps, and that the value of this factor needs to be adjusted for the
    specific problem. as an initial guideline, it is suggested that this be at
    least one to two orders of magnitude smaller than the solubility product.
    """

    @staticmethod
    def build_parameters(rblock, config):
        rblock.eps = Param(
            mutable=True, initialize=1e-4, doc="Smoothing parameter for smooth maximum"
        )

        rblock.s_norm = Param(
            mutable=True,
            initialize=1e-4,
            doc="Normalizing factor for solid precipitation term",
        )
        if hasattr(rblock, "k_eq_ref"):
            rblock.s_norm.set_value(value(rblock.k_eq_ref))

        rblock.s_scale = Param(
            mutable=True,
            initialize=10,
            doc="Scaling factor for solid precipitation term w.r.t saturated status Q = ln(Ksp) - ln(f(C))",
        )

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        e = None
        s = None

        if hasattr(b.params, "_electrolyte") and b.params._electrolyte:
            pc_set = b.params.true_phase_component_set
        else:
            pc_set = b.phase_component_set

        # Get reaction orders and construct power law expression
        for p, j in pc_set:
            p_obj = b.state_ref.params.get_phase(p)

            # First, build E
            o = rblock.reaction_order[p, j]

            if e is None and o.value != 0:
                e = get_concentration_term(b, r_idx, log=True)[p, j] * o
            elif e is not None and o.value != 0:
                e += get_concentration_term(b, r_idx, log=True)[p, j] * o

            if p_obj.is_solid_phase():
                # If solid phase, identify S
                r_config = b.params.config.equilibrium_reactions[r_idx]

                try:
                    stoic = r_config.stoichiometry[p, j]
                    if s is None and stoic != 0:
                        s = b.state_ref.flow_mol_phase_comp[p, j]
                    elif s is not None and stoic != 0:
                        s += b.state_ref.flow_mol_phase_comp[p, j]

                except KeyError:
                    pass

        if s is None:
            # Catch for not finding a solid phase
            raise ConfigurationError(
                "{} did not find a solid phase component for precipitation "
                "reaction {}. This is likely due to the reaction "
                "configuration.".format(b.name, r_idx)
            )
        else:
            # Need to remove units as complementarity is not consistent
            sunits = pyunits.get_units(s)

            if sunits is not None:
                s = s / sunits

        s = rblock.s_scale * s / (s + rblock.s_norm)

        Q = b.log_k_eq[r_idx] - e
        # Q should be unitless due to log form

        return Q - smooth_max(0, Q - s, rblock.eps) == 0

    @staticmethod
    def calculate_scaling_factors(b, sf_keq):
        return sf_keq



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
General purpose mixer block for IDAES models
"""
from enum import Enum

from pyomo.environ import (
    Block,
    check_optimal_termination,
    Constraint,
    Param,
    PositiveReals,
    Reals,
    RangeSet,
    Var,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf, Bool

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    MaterialBalanceType,
    MaterialFlowBasis,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_state_block,
)
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    PropertyNotSupportedError,
    InitializationError,
)
from idaes.core.util.math import smooth_min
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.initialization import ModularInitializerBase

import idaes.logger as idaeslog

__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


# Enumerate options for balances
class MixingType(Enum):
    """
    Enum of supported mixing types.
    """

    none = 0
    extensive = 1


class MomentumMixingType(Enum):
    """
    Enum of supported approaches to pressure mixing.
    """

    none = 0
    minimize = 1
    equality = 2
    minimize_and_equality = 3


class MixerInitializer(ModularInitializerBase):
    """
    Hierarchical Initializer for Mixer blocks.

    """

    def initialization_routine(
        self,
        model: Block,
    ):
        """
        Initialization routine for Mixer Blocks.

        This routine starts by initializing each of the inlet streams, then uses those
        results to estimate the outlet state before solving the full model.

        Args:
            model: model to be initialized

        Returns:
            Pyomo solver status object
        """
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        # Initialize inlet state blocks
        inlet_list = model.create_inlet_list()
        i_block_list = []
        for i in inlet_list:
            i_block = getattr(model, i + "_state")
            i_block_list.append(i_block)

            # Get initializer for inlet
            iinit = self.get_submodel_initializer(i_block)

            iinit.initialize(i_block)

        # Initialize mixed state block
        if model.config.mixed_state_block is None:
            mblock = model.mixed_state
        else:
            mblock = model.config.mixed_state_block

        # Calculate initial guesses for mixed stream state
        for t in model.flowsheet().time:
            # Iterate over state vars as defined by property package
            s_vars = mblock[t].define_state_vars()
            for s in s_vars:
                i_vars = []
                for k in s_vars[s]:
                    # If fixed, use current value
                    # otherwise calculate guess from mixed state
                    if not s_vars[s][k].fixed:
                        for ib in i_block_list:
                            i_vars.append(getattr(ib[t], s_vars[s].local_name))

                        if s == "pressure":
                            # If pressure, use minimum as initial guess
                            mblock[t].pressure.set_value(
                                min(
                                    i_block_list[i][t].pressure.value
                                    for i in range(len(i_block_list))
                                )
                            )
                        elif "flow" in s:
                            # If a "flow" variable (i.e. extensive), sum inlets
                            for k in s_vars[s]:
                                s_vars[s][k].set_value(
                                    sum(
                                        i_vars[i][k].value
                                        for i in range(len(i_block_list))
                                    )
                                )
                        else:
                            # Otherwise use average of inlets
                            for k in s_vars[s]:
                                s_vars[s][k].set_value(
                                    sum(
                                        i_vars[i][k].value
                                        for i in range(len(i_block_list))
                                    )
                                    / len(i_block_list)
                                )

        # Get initializer for mixed block
        minit = self.get_submodel_initializer(mblock)
        minit.initialize(mblock)

        res = None

        if model.config.mixed_state_block is None:
            if (
                hasattr(model, "pressure_equality_constraints")
                and model.pressure_equality_constraints.active is True
            ):
                model.pressure_equality_constraints.deactivate()
                for t in model.flowsheet().time:
                    sys_press = getattr(model, model.create_inlet_list()[0] + "_state")[
                        t
                    ].pressure
                    model.mixed_state[t].pressure.fix(sys_press.value)
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = solver.solve(model, tee=slc.tee)
                model.pressure_equality_constraints.activate()
                for t in model.flowsheet().time:
                    model.mixed_state[t].pressure.unfix()
            else:
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = solver.solve(model, tee=slc.tee)

            init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))
        else:
            init_log.info("Initialization Complete.")

        return res


@declare_process_block_class("Mixer")
class MixerData(UnitModelBlockData):
    """
    This is a general purpose model for a Mixer block with the IDAES modeling
    framework. This block can be used either as a stand-alone Mixer unit
    operation, or as a sub-model within another unit operation.

    This model creates a number of StateBlocks to represent the incoming
    streams, then writes a set of phase-component material balances, an
    overall enthalpy balance and a momentum balance (2 options) linked to a
    mixed-state StateBlock. The mixed-state StateBlock can either be specified
    by the user (allowing use as a sub-model), or created by the Mixer.

    When being used as a sub-model, Mixer should only be used when a set
    of new StateBlocks are required for the streams to be mixed. It should not
    be used to mix streams from multiple ControlVolumes in a single unit model -
    in these cases the unit model developer should write their own mixing
    equations.
    """

    default_initializer = MixerInitializer

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Mixer blocks are always steady-state.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Mixer blocks do not contain holdup, thus this must be
False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for mixer",
            doc="""Property parameter object used to define property
calculations, **default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "inlet_list",
        ConfigValue(
            domain=ListOf(str),
            description="List of inlet names",
            doc="""A list containing names of inlets,
**default** - None.
**Valid values:** {
**None** - use num_inlets argument,
**list** - a list of names to use for inlets.}""",
        ),
    )
    CONFIG.declare(
        "num_inlets",
        ConfigValue(
            domain=int,
            description="Number of inlets to unit",
            doc="""Argument indicating number (int) of inlets to construct, not
used if inlet_list arg is provided,
**default** - None.
**Valid values:** {
**None** - use inlet_list arg instead, or default to 2 if neither argument
provided,
**int** - number of inlets to create (will be named with sequential integers
from 1 to num_inlets).}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Calculate phase equilibrium in mixed stream",
            doc="""Argument indicating whether phase equilibrium should be
calculated for the resulting mixed stream,
**default** - False.
**Valid values:** {
**True** - calculate phase equilibrium in mixed stream,
**False** - do not calculate equilibrium in mixed stream.}""",
        ),
    )
    CONFIG.declare(
        "energy_mixing_type",
        ConfigValue(
            default=MixingType.extensive,
            domain=MixingType,
            description="Method to use when mixing energy flows",
            doc="""Argument indicating what method to use when mixing energy
flows of incoming streams,
**default** - MixingType.extensive.
**Valid values:** {
**MixingType.none** - do not include energy mixing equations,
**MixingType.extensive** - mix total enthalpy flows of each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_mixing_type",
        ConfigValue(
            default=MomentumMixingType.minimize,
            domain=MomentumMixingType,
            description="Method to use when mixing momentum/pressure",
            doc="""Argument indicating what method to use when mixing momentum/
pressure of incoming streams,
**default** - MomentumMixingType.minimize.
**Valid values:** {
**MomentumMixingType.none** - do not include momentum mixing equations,
**MomentumMixingType.minimize** - mixed stream has pressure equal to the
minimimum pressure of the incoming streams (uses smoothMin operator),
**MomentumMixingType.equality** - enforces equality of pressure in mixed and
all incoming streams.,
**MomentumMixingType.minimize_and_equality** - add constraints for pressure
equal to the minimum pressure of the inlets and constraints for equality of
pressure in mixed and all incoming streams. When the model is initially built,
the equality constraints are deactivated.  This option is useful for switching
between flow and pressure driven simulations.}""",
        ),
    )
    CONFIG.declare(
        "mixed_state_block",
        ConfigValue(
            default=None,
            domain=is_state_block,
            description="Existing StateBlock to use as mixed stream",
            doc="""An existing state block to use as the outlet stream from the
Mixer block,
**default** - None.
**Valid values:** {
**None** - create a new StateBlock for the mixed stream,
**StateBlock** - a StateBock to use as the destination for the mixed stream.}
""",
        ),
    )
    CONFIG.declare(
        "construct_ports",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Construct inlet and outlet Port objects",
            doc="""Argument indicating whether model should construct Port
objects linked to all inlet states and the mixed state,
**default** - True.
**Valid values:** {
**True** - construct Ports for all states,
**False** - do not construct Ports.""",
        ),
    )

    def build(self):
        """
        General build method for MixerData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        # Call super.build()
        super(MixerData, self).build()

        # Call setup methods from ControlVolumeBlockData
        self._get_property_package()
        self._get_indexing_sets()

        # Create list of inlet names
        inlet_list = self.create_inlet_list()

        # Build StateBlocks
        inlet_blocks = self.add_inlet_state_blocks(inlet_list)

        if self.config.mixed_state_block is None:
            mixed_block = self.add_mixed_state_block()
        else:
            mixed_block = self.get_mixed_state_block()

        mb_type = self.config.material_balance_type
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = mixed_block[t_ref].default_material_balance_type()

        if mb_type != MaterialBalanceType.none:
            self.add_material_mixing_equations(
                inlet_blocks=inlet_blocks, mixed_block=mixed_block, mb_type=mb_type
            )
        else:
            raise BurntToast(
                "{} received unrecognised value for "
                "material_mixing_type argument. This "
                "should not occur, so please contact "
                "the IDAES developers with this bug.".format(self.name)
            )

        if self.config.energy_mixing_type == MixingType.extensive:
            self.add_energy_mixing_equations(
                inlet_blocks=inlet_blocks, mixed_block=mixed_block
            )
        elif self.config.energy_mixing_type == MixingType.none:
            pass
        else:
            raise ConfigurationError(
                "{} received unrecognised value for "
                "material_mixing_type argument. This "
                "should not occur, so please contact "
                "the IDAES developers with this bug.".format(self.name)
            )

        # Add to try/expect to catch cases where pressure is not supported
        # by properties.
        try:
            if self.config.momentum_mixing_type == MomentumMixingType.minimize:
                self.add_pressure_minimization_equations(
                    inlet_blocks=inlet_blocks, mixed_block=mixed_block
                )
            elif self.config.momentum_mixing_type == MomentumMixingType.equality:
                self.add_pressure_equality_equations(
                    inlet_blocks=inlet_blocks, mixed_block=mixed_block
                )
            elif (
                self.config.momentum_mixing_type
                == MomentumMixingType.minimize_and_equality
            ):
                self.add_pressure_minimization_equations(
                    inlet_blocks=inlet_blocks, mixed_block=mixed_block
                )
                self.add_pressure_equality_equations(
                    inlet_blocks=inlet_blocks, mixed_block=mixed_block
                )
                self.pressure_equality_constraints.deactivate()
            elif self.config.momentum_mixing_type == MomentumMixingType.none:
                pass
            else:
                raise ConfigurationError(
                    "{} received unrecognised value for "
                    "momentum_mixing_type argument. This "
                    "should not occur, so please contact "
                    "the IDAES developers with this bug.".format(self.name)
                )
        except PropertyNotSupportedError:
            raise PropertyNotSupportedError(
                "{} The property package supplied for this unit does not "
                "appear to support pressure, which is required for momentum "
                "mixing. Please set momentum_mixing_type to "
                "MomentumMixingType.none or provide a property package which "
                "supports pressure.".format(self.name)
            )

        self.add_port_objects(inlet_list, inlet_blocks, mixed_block)

    def create_inlet_list(self):
        """
        Create list of inlet stream names based on config arguments.

        Returns:
            list of strings
        """
        if self.config.inlet_list is not None and self.config.num_inlets is not None:
            # If both arguments provided and not consistent, raise Exception
            if len(self.config.inlet_list) != self.config.num_inlets:
                raise ConfigurationError(
                    "{} Mixer provided with both inlet_list and "
                    "num_inlets arguments, which were not consistent ("
                    "length of inlet_list was not equal to num_inlets). "
                    "PLease check your arguments for consistency, and "
                    "note that it is only necessary to provide one of "
                    "these arguments.".format(self.name)
                )
        elif self.config.inlet_list is None and self.config.num_inlets is None:
            # If no arguments provided for inlets, default to num_inlets = 2
            self.config.num_inlets = 2

        # Create a list of names for inlet StateBlocks
        if self.config.inlet_list is not None:
            inlet_list = self.config.inlet_list
        else:
            inlet_list = [
                "inlet_" + str(n) for n in range(1, self.config.num_inlets + 1)
            ]

        return inlet_list

    def add_inlet_state_blocks(self, inlet_list):
        """
        Construct StateBlocks for all inlet streams.

        Args:
            list of strings to use as StateBlock names

        Returns:
            list of StateBlocks
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        # Create empty list to hold StateBlocks for return
        inlet_blocks = []

        # Create an instance of StateBlock for all inlets
        for i in inlet_list:
            i_obj = self.config.property_package.build_state_block(
                self.flowsheet().time, doc="Material properties at inlet", **tmp_dict
            )

            setattr(self, i + "_state", i_obj)

            inlet_blocks.append(getattr(self, i + "_state"))

        return inlet_blocks

    def add_mixed_state_block(self):
        """
        Constructs StateBlock to represent mixed stream.

        Returns:
            New StateBlock object
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = self.config.has_phase_equilibrium
        tmp_dict["defined_state"] = False

        self.mixed_state = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties of mixed stream", **tmp_dict
        )

        return self.mixed_state

    def get_mixed_state_block(self):
        """
        Validates StateBlock provided in user arguments for mixed stream.

        Returns:
            The user-provided StateBlock or an Exception
        """
        # Sanity check to make sure method is not called when arg missing
        if self.config.mixed_state_block is None:
            raise BurntToast(
                "{} get_mixed_state_block method called when "
                "mixed_state_block argument is None. This should "
                "not happen.".format(self.name)
            )

        # Check that the user-provided StateBlock uses the same prop pack
        if (
            self.config.mixed_state_block[
                self.flowsheet().time.first()
            ].config.parameters
            != self.config.property_package
        ):
            raise ConfigurationError(
                "{} StateBlock provided in mixed_state_block argument "
                "does not come from the same property package as "
                "provided in the property_package argument. All "
                "StateBlocks within a Mixer must use the same "
                "property package.".format(self.name)
            )

        return self.config.mixed_state_block

    def add_material_mixing_equations(self, inlet_blocks, mixed_block, mb_type):
        """
        Add material mixing equations.
        """
        pp = self.config.property_package
        # Get phase component list(s)
        pc_set = mixed_block.phase_component_set

        # Get units metadata
        units = pp.get_metadata()

        flow_basis = mixed_block[
            self.flowsheet().time.first()
        ].get_material_flow_basis()
        if flow_basis == MaterialFlowBasis.molar:
            flow_units = units.get_derived_units("flow_mole")
        elif flow_basis == MaterialFlowBasis.mass:
            flow_units = units.get_derived_units("flow_mass")
        else:
            # Let this pass for now with no units
            flow_units = None

        if mixed_block.include_inherent_reactions:
            if mb_type == MaterialBalanceType.total:
                raise ConfigurationError(
                    "Cannot do total flow mixing with inherent reaction; "
                    "problem is under-constrained. Please use a different "
                    "mixing type."
                )

            # Add extents of reaction and stoichiometric constraints
            self.inherent_reaction_extent = Var(
                self.flowsheet().time,
                mixed_block.params.inherent_reaction_idx,
                domain=Reals,
                initialize=0.0,
                doc="Extent of inherent reactions in outlet",
                units=flow_units,
            )

            self.inherent_reaction_generation = Var(
                self.flowsheet().time,
                pc_set,
                domain=Reals,
                initialize=0.0,
                doc="Generation due to inherent reactions in outlet",
                units=flow_units,
            )

            @self.Constraint(
                self.flowsheet().time,
                pc_set,
            )
            def inherent_reaction_constraint(b, t, p, j):
                if (p, j) in pc_set:
                    return b.inherent_reaction_generation[t, p, j] == (
                        sum(
                            mixed_block[t].params.inherent_reaction_stoichiometry[
                                r, p, j
                            ]
                            * self.inherent_reaction_extent[t, r]
                            for r in mixed_block[t].params.inherent_reaction_idx
                        )
                    )
                return Constraint.Skip

        if mb_type == MaterialBalanceType.componentPhase:
            # Create equilibrium generation term and constraints if required
            if self.config.has_phase_equilibrium is True:
                try:
                    self.phase_equilibrium_generation = Var(
                        self.flowsheet().time,
                        pp.phase_equilibrium_idx,
                        domain=Reals,
                        doc="Amount of generation in unit by phase equilibria",
                        units=flow_units,
                    )
                except AttributeError:
                    raise PropertyNotSupportedError(
                        "{} Property package does not contain a list of phase "
                        "equilibrium reactions (phase_equilibrium_idx), "
                        "thus does not support phase equilibrium.".format(self.name)
                    )

            # Write phase-component balances
            @self.Constraint(
                self.flowsheet().time,
                pc_set,
                doc="Material mixing equations",
            )
            def material_mixing_equations(b, t, p, j):
                rhs = sum(
                    inlet_blocks[i][t].get_material_flow_terms(p, j)
                    for i in range(len(inlet_blocks))
                ) - mixed_block[t].get_material_flow_terms(p, j)

                if self.config.has_phase_equilibrium:
                    rhs += sum(
                        b.phase_equilibrium_generation[t, r]
                        for r in pp.phase_equilibrium_idx
                        if pp.phase_equilibrium_list[r][0] == j
                        and pp.phase_equilibrium_list[r][1][0] == p
                    ) - sum(
                        b.phase_equilibrium_generation[t, r]
                        for r in pp.phase_equilibrium_idx
                        if pp.phase_equilibrium_list[r][0] == j
                        and pp.phase_equilibrium_list[r][1][1] == p
                    )

                if mixed_block.include_inherent_reactions:
                    rhs += b.inherent_reaction_generation[t, p, j]

                return 0 == rhs

        elif mb_type == MaterialBalanceType.componentTotal:
            # Write phase-component balances
            @self.Constraint(
                self.flowsheet().time,
                mixed_block.component_list,
                doc="Material mixing equations",
            )
            def material_mixing_equations(b, t, j):
                rhs = sum(
                    sum(
                        inlet_blocks[i][t].get_material_flow_terms(p, j)
                        for i in range(len(inlet_blocks))
                    )
                    - mixed_block[t].get_material_flow_terms(p, j)
                    for p in mixed_block.phase_list
                    if (p, j) in pc_set
                )

                if mixed_block.include_inherent_reactions:
                    rhs += sum(
                        b.inherent_reaction_generation[t, p, j]
                        for p in mixed_block.phase_list
                        if (p, j) in pc_set
                    )

                return 0 == rhs

        elif mb_type == MaterialBalanceType.total:
            # Write phase-component balances
            @self.Constraint(self.flowsheet().time, doc="Material mixing equations")
            def material_mixing_equations(b, t):
                rhs = sum(
                    sum(
                        sum(
                            inlet_blocks[i][t].get_material_flow_terms(p, j)
                            for i in range(len(inlet_blocks))
                        )
                        - mixed_block[t].get_material_flow_terms(p, j)
                        for j in mixed_block.component_list
                        if (p, j) in pc_set
                    )
                    for p in mixed_block.phase_list
                )

                return 0 == rhs

        elif mb_type == MaterialBalanceType.elementTotal:
            raise ConfigurationError(
                "{} Mixers do not support elemental "
                "material balances.".format(self.name)
            )
        elif mb_type == MaterialBalanceType.none:
            pass
        else:
            raise BurntToast(
                "{} Mixer received unrecognised value for "
                "material_balance_type. This should not happen, "
                "please report this bug to the IDAES developers.".format(self.name)
            )

    def add_energy_mixing_equations(self, inlet_blocks, mixed_block):
        """
        Add energy mixing equations (total enthalpy balance).
        """

        @self.Constraint(self.flowsheet().time, doc="Energy balances")
        def enthalpy_mixing_equations(b, t):
            return 0 == (
                sum(
                    sum(
                        inlet_blocks[i][t].get_enthalpy_flow_terms(p)
                        for p in mixed_block.phase_list
                    )
                    for i in range(len(inlet_blocks))
                )
                - sum(
                    mixed_block[t].get_enthalpy_flow_terms(p)
                    for p in mixed_block.phase_list
                )
            )

    def add_pressure_minimization_equations(self, inlet_blocks, mixed_block):
        """
        Add pressure minimization equations. This is done by sequential
        comparisons of each inlet to the minimum pressure so far, using
        the IDAES smooth minimum function.
        """
        if not hasattr(self, "inlet_idx"):
            self.inlet_idx = RangeSet(len(inlet_blocks))

        # Get units metadata
        units = mixed_block.params.get_metadata()

        # Add variables
        self.minimum_pressure = Var(
            self.flowsheet().time,
            self.inlet_idx,
            doc="Variable for calculating minimum inlet pressure",
            units=units.get_derived_units("pressure"),
        )

        self.eps_pressure = Param(
            mutable=True,
            initialize=1e-3,
            domain=PositiveReals,
            doc="Smoothing term for minimum inlet pressure",
            units=units.get_derived_units("pressure"),
        )

        # Calculate minimum inlet pressure
        @self.Constraint(
            self.flowsheet().time,
            self.inlet_idx,
            doc="Calculation for minimum inlet pressure",
        )
        def minimum_pressure_constraint(b, t, i):
            if i == self.inlet_idx.first():
                return self.minimum_pressure[t, i] == (inlet_blocks[i - 1][t].pressure)
            else:
                return self.minimum_pressure[t, i] == (
                    smooth_min(
                        self.minimum_pressure[t, i - 1],
                        inlet_blocks[i - 1][t].pressure,
                        self.eps_pressure,
                    )
                )

        # Set inlet pressure to minimum pressure
        @self.Constraint(self.flowsheet().time, doc="Link pressure to control volume")
        def mixture_pressure(b, t):
            return mixed_block[t].pressure == (
                self.minimum_pressure[t, self.inlet_idx.last()]
            )

    def add_pressure_equality_equations(self, inlet_blocks, mixed_block):
        """
        Add pressure equality equations. Note that this writes a number of
        constraints equal to the number of inlets, enforcing equality between
        all inlets and the mixed stream.
        """
        if not hasattr(self, "inlet_idx"):
            self.inlet_idx = RangeSet(len(inlet_blocks))

        # Create equality constraints
        @self.Constraint(
            self.flowsheet().time,
            self.inlet_idx,
            doc="Calculation for minimum inlet pressure",
        )
        def pressure_equality_constraints(b, t, i):
            return mixed_block[t].pressure == inlet_blocks[i - 1][t].pressure

    def add_port_objects(self, inlet_list, inlet_blocks, mixed_block):
        """
        Adds Port objects if required.

        Args:
            a list of inlet StateBlock objects
            a mixed state StateBlock object

        Returns:
            None
        """
        if self.config.construct_ports is True:
            # Add ports
            for p in inlet_list:
                i_state = getattr(self, p + "_state")
                self.add_port(name=p, block=i_state, doc="Inlet Port")
            self.add_port(name="outlet", block=mixed_block, doc="Outlet Port")

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().time:
            try:
                inlet_list = blk.create_inlet_list()
                for i in inlet_list:
                    i_block = getattr(blk, i + "_state")
                    i_block[t].model_check()
            except AttributeError:
                _log.warning(
                    "{} Mixer inlet property block has no model "
                    "checks. To correct this, add a model_check "
                    "method to the associated StateBlock class.".format(blk.name)
                )
            try:
                if blk.config.mixed_state_block is None:
                    blk.mixed_state[t].model_check()
                else:
                    blk.config.mixed_state_block.model_check()
            except AttributeError:
                _log.warning(
                    "{} Mixer outlet property block has no "
                    "model checks. To correct this, add a "
                    "model_check method to the associated "
                    "StateBlock class.".format(blk.name)
                )

    def use_minimum_inlet_pressure_constraint(self):
        """Activate the mixer pressure = minimum inlet pressure constraint and
        deactivate the mixer pressure and all inlet pressures are equal
        constraints. This should only be used when momentum_mixing_type ==
        MomentumMixingType.minimize_and_equality.
        """
        if self.config.momentum_mixing_type != MomentumMixingType.minimize_and_equality:
            _log.warning(
                """use_minimum_inlet_pressure_constraint() can only be used
                when momentum_mixing_type ==
                MomentumMixingType.minimize_and_equality"""
            )
            return
        self.minimum_pressure_constraint.activate()
        self.pressure_equality_constraints.deactivate()

    def use_equal_pressure_constraint(self):
        """Deactivate the mixer pressure = minimum inlet pressure constraint
        and activate the mixer pressure and all inlet pressures are equal
        constraints. This should only be used when momentum_mixing_type ==
        MomentumMixingType.minimize_and_equality.
        """
        if self.config.momentum_mixing_type != MomentumMixingType.minimize_and_equality:
            _log.warning(
                """use_equal_pressure_constraint() can only be used when
                momentum_mixing_type ==
                MomentumMixingType.minimize_and_equality"""
            )
            return
        self.minimum_pressure_constraint.deactivate()
        self.pressure_equality_constraints.activate()

    def fix_initialization_states(self):
        """
        Iterate over inlet ports and fix all variables.

        For Mixers with pressure equality, we will assume that pressure has been
        correctly specified and not fix pressures.

        Returns:
            None
        """
        inlet_list = self.create_inlet_list()
        for p in inlet_list:
            p_obj = getattr(self, p)
            # Iterate over vars
            for v in p_obj.iter_vars():
                if (
                    self.config.momentum_mixing_type == MomentumMixingType.equality
                    or self.config.momentum_mixing_type
                    == MomentumMixingType.minimize_and_equality
                ) and v.local_name == "pressure":
                    # Don't fix pressure in cases where pressure equality is specified
                    continue
                else:
                    v.fix()

    def initialize_build(
        blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None, hold_state=False
    ):
        """
        Initialization routine for mixer.

        Keyword Arguments:
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # Initialize inlet state blocks
        flags = {}
        inlet_list = blk.create_inlet_list()
        i_block_list = []
        for i in inlet_list:
            i_block = getattr(blk, i + "_state")
            i_block_list.append(i_block)
            flags[i] = {}
            flags[i] = i_block.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                hold_state=True,
            )

        # Initialize mixed state block
        if blk.config.mixed_state_block is None:
            mblock = blk.mixed_state
        else:
            mblock = blk.config.mixed_state_block

        o_flags = {}
        # Calculate initial guesses for mixed stream state
        for t in blk.flowsheet().time:
            # Iterate over state vars as defined by property package
            s_vars = mblock[t].define_state_vars()
            for s in s_vars:
                i_vars = []
                for k in s_vars[s]:
                    # Record whether variable was fixed or not
                    o_flags[t, s, k] = s_vars[s][k].fixed

                    # If fixed, use current value
                    # otherwise calculate guess from mixed state
                    if not s_vars[s][k].fixed:
                        for ib in i_block_list:
                            i_vars.append(getattr(ib[t], s_vars[s].local_name))

                        if s == "pressure":
                            # If pressure, use minimum as initial guess
                            mblock[t].pressure.value = min(
                                i_block_list[i][t].pressure.value
                                for i in range(len(i_block_list))
                            )
                        elif "flow" in s:
                            # If a "flow" variable (i.e. extensive), sum inlets
                            for k in s_vars[s]:
                                s_vars[s][k].value = sum(
                                    i_vars[i][k].value for i in range(len(i_block_list))
                                )
                        else:
                            # Otherwise use average of inlets
                            for k in s_vars[s]:
                                s_vars[s][k].value = sum(
                                    i_vars[i][k].value for i in range(len(i_block_list))
                                ) / len(i_block_list)

        mblock.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=False,
        )

        res = None
        # Revert fixed status of variables to what they were before
        for t in blk.flowsheet().time:
            s_vars = mblock[t].define_state_vars()
            for s in s_vars:
                for k in s_vars[s]:
                    s_vars[s][k].fixed = o_flags[t, s, k]

        if blk.config.mixed_state_block is None:
            if (
                hasattr(blk, "pressure_equality_constraints")
                and blk.pressure_equality_constraints.active is True
            ):
                blk.pressure_equality_constraints.deactivate()
                for t in blk.flowsheet().time:
                    sys_press = getattr(blk, blk.create_inlet_list()[0] + "_state")[
                        t
                    ].pressure
                    blk.mixed_state[t].pressure.fix(sys_press.value)
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = opt.solve(blk, tee=slc.tee)
                blk.pressure_equality_constraints.activate()
                for t in blk.flowsheet().time:
                    blk.mixed_state[t].pressure.unfix()
            else:
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = opt.solve(blk, tee=slc.tee)

            init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))
        else:
            init_log.info("Initialization Complete.")

        if res is not None and not check_optimal_termination(res):
            raise InitializationError(
                f"{blk.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags, outlvl=outlvl)

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        """
        inlet_list = blk.create_inlet_list()
        for i in inlet_list:
            i_block = getattr(blk, i + "_state")
            i_block.release_state(flags[i], outlvl=outlvl)

    def _get_stream_table_contents(self, time_point=0):
        io_dict = {}
        inlet_list = self.create_inlet_list()
        for i in inlet_list:
            io_dict[i] = getattr(self, i + "_state")
        if self.config.mixed_state_block is None:
            io_dict["Outlet"] = self.mixed_state
        else:
            io_dict["Outlet"] = self.config.mixed_state_block
        return create_stream_table_dataframe(io_dict, time_point=time_point)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        mb_type = self.config.material_balance_type
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = self.mixed_state[t_ref].default_material_balance_type()

        if hasattr(self, "pressure_equality_constraints"):
            for (t, i), c in self.pressure_equality_constraints.items():
                s = iscale.get_scaling_factor(
                    self.mixed_state[t].pressure, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "minimum_pressure"):
            for (t, i), v in self.minimum_pressure.items():
                s = iscale.get_scaling_factor(
                    self.mixed_state[t].pressure, default=1, warning=True
                )
                iscale.set_scaling_factor(v, s)

        if hasattr(self, "minimum_pressure_constraint"):
            for (t, i), c in self.minimum_pressure_constraint.items():
                s = iscale.get_scaling_factor(
                    self.mixed_state[t].pressure, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "mixture_pressure"):
            for t, c in self.mixture_pressure.items():
                s = iscale.get_scaling_factor(
                    self.mixed_state[t].pressure, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "material_mixing_equations"):
            if mb_type == MaterialBalanceType.componentPhase:
                for (t, p, j), c in self.material_mixing_equations.items():
                    flow_term = self.mixed_state[t].get_material_flow_terms(p, j)
                    s = iscale.get_scaling_factor(flow_term, default=1)
                    iscale.constraint_scaling_transform(c, s, overwrite=False)
            elif mb_type == MaterialBalanceType.componentTotal:
                for (t, j), c in self.material_mixing_equations.items():
                    for i, p in enumerate(self.mixed_state.phase_list):
                        try:
                            ft = self.mixed_state[t].get_material_flow_terms(p, j)
                        except (KeyError, AttributeError):
                            continue  # component not in phase
                        if i == 0:
                            s = iscale.get_scaling_factor(ft, default=1)
                        else:
                            _s = iscale.get_scaling_factor(ft, default=1)
                            s = _s if _s < s else s
                    iscale.constraint_scaling_transform(c, s, overwrite=False)
            elif mb_type == MaterialBalanceType.total:
                pc_set = self.mixed_state.phase_component_set
                for t, c in self.material_mixing_equations.items():
                    for i, (p, j) in enumerate(pc_set):
                        ft = self.mixed_state[t].get_material_flow_terms(p, j)
                        if i == 0:
                            s = iscale.get_scaling_factor(ft, default=1)
                        else:
                            _s = iscale.get_scaling_factor(ft, default=1)
                            s = _s if _s < s else s
                    iscale.constraint_scaling_transform(c, s, overwrite=False)

        if hasattr(self, "enthalpy_mixing_equations"):

            def scale_gen(t):
                for p in self.mixed_state[t].phase_list:
                    yield self.mixed_state[t].get_enthalpy_flow_terms(p)

            for t, c in self.enthalpy_mixing_equations.items():
                s = iscale.min_scaling_factor(scale_gen(t), default=1)
                iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Generate parameter and expression files for r1234ze(e)
"""

__author__ = "John Eslick"

import os
from pyomo.common.fileutils import this_file_dir
from idaes.models.properties.general_helmholtz.helmholtz_parameters import (
    WriteParameters,
)


def thermal_conductivity_rule(m):
    """Thermal Conductivity

    Richard A. Perkins and Marcia L. Huber. Measurement and Correlation of the Thermal
        Conductivity of 2,3,3,3-Tetrafluoroprop-1-ene (R1234yf) and
        trans-1,3,3,3-Tetrafluoropropene (R1234ze(E)). J. Chem. Eng. Data, 56:4868–4874,
        2011. doi:10.1021/je200811n.
    """
    a = [
        -0.0103589,
        0.0308929,
        0.000230348,
    ]
    b = {
        (1, 1): -0.0428296,
        (1, 2): 0.0434288,
        (2, 1): 0.0927099,
        (2, 2): -0.0605844,
        (3, 1): -0.0702107,
        (3, 2): 0.0440187,
        (4, 1): 0.0249708,
        (4, 2): -0.0155082,
        (5, 1): -0.00301838,
        (5, 2): 0.00210190,
    }
    Tred = 1.0 / m.tau
    l0 = sum(ai * Tred**i for i, ai in enumerate(a))
    lr = sum(m.delta**i * (b[i, 1] + b[i, 2] * Tred) for i in range(1, 6))
    lc = 0
    return 1000 * (l0 + lr + lc)


def viscosity_rule(m):
    """Viscosity rule

    Huber ML, Assael MJ. Correlations for the Viscosity of 2,3,3,3-Tetrafluoroprop-1-ene
        (R1234yf) and trans-1,3,3,3-Tetrafluoropropene (R1234ze(E)). Int J Refrig.
        2016; 71:39-45. doi:10.1016/j.ijrefrig.2016.08.007
    """

    a = [
        -963382,
        9614.09,
        -13.233,
        0.0360562,
        122059,
        -224.741,
    ]
    b = [
        -19.572881,
        219.73999,
        -1015.3226,
        2471.0125,
        -3375.1717,
        2491.6597,
        -787.26086,
        14.085455,
        -0.34664158,
    ]
    c = [
        8.61691913,
        0,
        20.83024738,
        0,
        0.54243690,
        -10.49684841,
        -1.38137689,
        1,
        0,
    ]
    T = m.T_star / m.tau
    rho = m.delta * m.rho_star / m.MW * 1000
    eok = 340
    sigma = 5.017e-1
    NA = 6.0221408e23
    Ts = T / eok
    eta0 = (a[0] + a[1] * T + a[2] * T**2 + a[3] * T**3) / (a[4] + a[5] * T + T**2)
    Bs = (
        sum(b[i] * Ts ** (-0.25 * i) for i in range(0, 7))
        + b[7] * Ts**-2.5
        + b[8] * Ts**-5.5
    )
    B = NA * Bs * sigma**3 / 1e9**3
    eta1 = eta0 * B
    Tr = 1 / m.tau
    etar = (
        m.delta ** (2.0 / 3.0)
        * Tr**0.5
        * (
            c[0]
            + c[1] * m.delta
            + c[2] * m.delta**2
            + (
                c[3] * m.delta
                + c[4] * m.delta**6
                + c[5] * m.delta * Tr**2
                + c[6] * m.delta**5 * Tr
            )
            / (c[7] * Tr + c[8] * m.delta * Tr)
        )
    )
    return eta0 + eta1 * rho + etar


def main(dry_run=False):
    """Generate parameter and expression files.

    Args:
        dry_run (bool): If dry run don't generate files

    Returns:
        None
    """
    main_param_file = os.path.join(this_file_dir(), "r1234ze.json")
    we = WriteParameters(parameters=main_param_file)

    we.add(
        {
            "viscosity": viscosity_rule,
            "thermal_conductivity": thermal_conductivity_rule,
        }
    )
    we.write(dry_run=dry_run)

    print("ASHRAE Offset")
    print(we.calculate_reference_offset(2.7584034882, 1.64063049539, 0, 0))
    return we


if __name__ == "__main__":
    main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

import logging
from contextlib import contextmanager

from pyomo.common.tee import capture_output
from pyomo.common.log import LogStream

import idaes

# Throw the standard levels in here, just let you access it all in one place
CRITICAL = logging.CRITICAL  # 50
ERROR = logging.ERROR  # 40
WARNING = logging.WARNING  # 30
CAUTION = 25
INFO_LOW = 21  # Most important info
INFO = logging.INFO  # 20  #Medium info (default)
INFO_HIGH = 19  # Less important important info
DEBUG = logging.DEBUG  # 10
NOTSET = logging.NOTSET  # 0


levelname = {
    CAUTION: "CAUTION",
    INFO_HIGH: "INFO",
    INFO_LOW: "INFO",
}


class _TagFilter(logging.Filter):
    """Filter applied to IDAES loggers returned by this module."""

    @staticmethod
    def filter(record):
        """Add in the custom level name and let the record through"""
        if record.levelno in levelname:
            record.levelname = levelname[record.levelno]
        if record.levelno >= WARNING:
            return True
        try:
            if record.tag is None or record.tag in idaes.cfg.logger_tags:
                return True
        except AttributeError:
            return True
        return False


def __caution(self, *args, **kwargs):
    self.log(CAUTION, *args, **kwargs)


def __info_low(self, *args, **kwargs):
    self.log(INFO_LOW, *args, **kwargs)


def __info_high(self, *args, **kwargs):
    self.log(INFO_HIGH, *args, **kwargs)


def __add_methods(log, tag=None):
    # pylint: disable=assignment-from-no-return,no-value-for-parameter
    log.addFilter(_TagFilter)
    log = logging.LoggerAdapter(log, {"tag": tag})
    log.caution = __caution.__get__(log)
    log.info_high = __info_high.__get__(log)
    log.info_low = __info_low.__get__(log)
    # hopefully adding this multiple times is not a problem
    return log


def _getLogger(name, logger_name="idaes", level=None, tag=None):
    assert tag in idaes.cfg.valid_logger_tags.union({None})
    if name.startswith("idaes."):
        name = name[6:]
    name = ".".join([logger_name, name])
    l = logging.getLogger(name)
    if level is not None:
        l.setLevel(level)
    return __add_methods(logging.getLogger(name), tag)


def getIdaesLogger(name, level=None, tag=None):
    """Return an idaes logger.

    Args:
        name: usually __name__
        level: standard IDAES logging level (default use IDAES config)
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes", level=level, tag=tag)


getLogger = getIdaesLogger


def getSolveLogger(name, level=None, tag=None):
    """Get a solver logger

    Args:
        name: logger name is "idaes.solve." + name (if name starts with "idaes."
            it is removed before creating the logger name)
        level: Log level
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes.solve", level=level, tag=tag)


def getInitLogger(name, level=None, tag=None):
    """Get a model initialization logger

    Args:
        name: Object name (usually Pyomo Component name)
        level: Log level
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes.init", level=level, tag=tag)


def getModelLogger(name, level=None, tag=None):
    """Get a logger for an IDAES model. This function helps users keep their
    loggers in a standard location and use the IDAES logging config.

    Args:
        name: Name (usually __name__).  Any starting 'idaes.' is stripped off, so
            if a model is part of the idaes package, 'idaes' won't be repeated.
        level: Standard Python logging level (default use IDAES config)
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes.model", level=level, tag=tag)


def condition(res):
    """Get the solver termination condition to log.  This isn't a specific value
    that you can really depend on, just a message to pass on from the solver for
    the user's benefit. Sometimes the solve is in a try-except, so we'll handle
    None and str for those cases, where you don't have a real result."""

    if res is None:
        return "Error, no result"
    elif isinstance(res, str):
        return res
    else:
        s = str(res.solver.termination_condition)

    try:
        if "ipopt" in str(res.solver.message).lower():
            solver_message = " ".join(str(res.solver.message).split(" ")[2:])
            return f"{s} - {solver_message}"
        else:
            return f"{s} - {str(res.solver.message)}"
    except:  # pylint: disable=bare-except
        return s


def solver_capture_on():
    """This function turns on the solver capture for the solver_log context
    manager. If this is on, solver output within the solver_log context
    is captured and sent to the logger.

    """
    idaes.cfg.logger_capture_solver = True


def solver_capture_off():
    """This function turns off the solver capture for the solver_log context
    manager. If this is off, solver output within the solver_log context
    is just sent to stdout like normal.

    """
    idaes.cfg.logger_capture_solver = False


def solver_capture():
    """Return True if solver capture is on or False otherwise."""
    return idaes.cfg.logger_capture_solver


def log_tags():
    """Returns a set of logging tags to be logged.

    Returns:
        (set) tags to be logged
    """
    return idaes.cfg.logger_tags


def set_log_tags(tags):
    """Specify a set of tags to be logged

    Args:
        tags(iterable of str): Tags to log

    Returns:
        None
    """
    for m in tags:
        if m not in idaes.cfg.valid_logger_tags.union({None}):
            raise ValueError(f"{m} is not a valid logging tag")
    idaes.cfg.logger_tags = set(tags)


def add_log_tag(tag):
    """Add a tag to the list of tags to log.

    Args:
        tag(str): Tag to log

    Returns:
        None
    """
    if tag not in idaes.cfg.valid_logger_tags.union({None}):
        raise ValueError(f"{tag} is not a valid logging tag")
    idaes.cfg.logger_tags.add(tag)


def remove_log_tag(tag):
    """Remove a tag from the list of tags to log.

    Args:
        tag(str): Tag to no longer log

    Returns:
        None
    """
    try:
        idaes.cfg.logger_tags.remove(tag)
    except ValueError:
        pass


def valid_log_tags():
    """Returns a set of valid logging tag names.

    Returns:
        (set) valid tag names
    """
    return idaes.cfg.valid_logger_tags.union({None})


def add_valid_log_tag(tag):
    """Add a tag name to the list of valid names.

    Args:
        tag(str): A tag name

    Returns:
        None
    """
    assert isinstance(tag, str)
    idaes.cfg.valid_logger_tags.add(tag)


class SolverLogInfo(object):
    def __init__(self, tee=True):
        self.tee = tee


@contextmanager
def solver_log(logger, level=logging.ERROR):
    """Context manager to send solver output to a logger."""
    tee = logger.isEnabledFor(level)
    if not solver_capture():
        yield SolverLogInfo(tee=tee)
    else:
        with capture_output(LogStream(level, logger)):
            yield SolverLogInfo(tee=tee)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Library of common forms for phase equilibrium constraints
"""
from pyomo.environ import ConcreteModel, Expression, value, Var, units as pyunits

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx

from idaes.models.properties.modular_properties.phase_equil.henry import (
    ConstantH,
    HenryType,
    henry_equilibrium_ratio,
    _raise_henry_type_error,
    get_henry_concentration_term,
)
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity
from pytest import approx
import pytest
from idaes.core.util.exceptions import ConfigurationError


# Dummy EoS
class DummyEoS(object):
    def common(self, *args, **kwargs):
        pass

    def build_parameters(b):
        pass

    def fug_phase_comp_eq(b, p, j, pp):
        return b.x[p, j]

    def log_fug_phase_comp_eq(b, p, j, pp):
        return 42 * b.x[p, j]


@pytest.mark.unit
def test_henry_invalid_phase_type():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    with pytest.raises(
        ConfigurationError,
        match="params component H2O was marked as a Henry's "
        "Law component in phase Vap, but this is not a Liquid "
        "phase.",
    ):
        m.params = GenericParameterBlock(
            components={
                "H2O": {
                    "parameter_data": {"temperature_crit": 647.3},
                    "henry_component": {"Vap": ConstantH},
                    "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                }
            },
            phases={
                "Liq": {"equation_of_state": DummyEoS},
                "Vap": {"equation_of_state": DummyEoS},
            },
            state_definition=FTPx,
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )


@pytest.mark.unit
def test_henry_invalid_phase_name():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    with pytest.raises(
        ConfigurationError,
        match="params component H2O was marked as a Henry's "
        "Law component in phase foo, but this is not a valid "
        "phase name.",
    ):
        m.params = GenericParameterBlock(
            components={
                "H2O": {
                    "parameter_data": {"temperature_crit": 647.3},
                    "henry_component": {"foo": ConstantH},
                    "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                }
            },
            phases={
                "Liq": {"equation_of_state": DummyEoS},
                "Vap": {"equation_of_state": DummyEoS},
            },
            state_definition=FTPx,
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
        )


@pytest.mark.unit
def test_constant_H():
    m = ConcreteModel()

    # Add a dummy var for use in constructing expressions
    m.x = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    m.mole_frac_phase_comp = Var(["Vap", "Liq"], ["H2O"], initialize=1)

    # Create a dummy parameter block
    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3, "henry_ref": {"Liq": 86}},
                "henry_component": {
                    "Liq": {"method": ConstantH, "type": HenryType.Kpx}
                },
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {"equation_of_state": DummyEoS},
            "Vap": {"equation_of_state": DummyEoS},
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    )

    assert isinstance(m.params.H2O.henry_ref_Liq, Var)
    assert m.params.H2O.henry_ref_Liq.value == 86

    m.state = m.params.build_state_block([0])

    H = ConstantH.return_expression(m.state[0], "Liq", "H2O")
    assert H is m.params.H2O.henry_ref_Liq

    assert isinstance(m.state[0].henry, Expression)
    assert len(m.state[0].henry) == 1
    assert m.state[0].henry["Liq", "H2O"].expr is m.params.H2O.henry_ref_Liq


@pytest.mark.unit
def test_invalid_henry_type():
    with pytest.raises(
        ConfigurationError, match="Unrecognized value for HenryType HenryType.Dummy"
    ):
        _raise_henry_type_error(HenryType.Dummy)

    m = ConcreteModel()
    m.henry = {("Liq", "H2O"): 1}
    henry_dict = {"type": HenryType.Dummy}

    def debugMethod():
        return 42

    m.params = GenericParameterBlock(
        components={
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3, "henry_ref": {"Liq": 86}},
                "henry_component": {
                    "Liq": {"method": debugMethod, "type": HenryType.Dummy}
                },
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        phases={
            "Liq": {"equation_of_state": DummyEoS},
            "Vap": {"equation_of_state": DummyEoS},
        },
        state_definition=FTPx,
        pressure_ref=100000.0,
        temperature_ref=300,
        base_units={
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    )

    with pytest.raises(
        ConfigurationError, match="Unrecognized value for HenryType HenryType.Dummy"
    ):
        get_henry_concentration_term(m, henry_dict)

    with pytest.raises(
        ConfigurationError, match="Unrecognized value for HenryType HenryType.Dummy"
    ):
        henry_equilibrium_ratio(m, "Liq", "H2O")


@pytest.mark.component
def test_equilibrium_ratio():
    config_dict = {
        "components": {
            "H2O": {
                "parameter_data": {"temperature_crit": 647.3, "henry_ref": {"Liq": 7}},
                "henry_component": {"Liq": {"method": ConstantH, "type": None}},
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            }
        },
        "phases": {
            "Liq": {"equation_of_state": DummyEoS},
            "Vap": {"equation_of_state": DummyEoS},
        },
        "state_definition": FTPx,
        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
    }

    henry_types = [HenryType.Hcp, HenryType.Kpc, HenryType.Hxp, HenryType.Kpx]
    expected_split = [3 / (5 * 7), (3 * 7) / 5, 1 / (5 * 7), 7 / 5]

    for htype, split in zip(henry_types, expected_split):
        config_dict["components"]["H2O"]["henry_component"]["Liq"]["type"] = htype
        m = ConcreteModel()
        m.params = GenericParameterBlock(**config_dict)

        m.state = m.params.build_state_block([0])
        m.state[0].mole_frac_phase_comp["Liq", "H2O"].value = 0.5
        m.state[0].dens_mol_phase = Var(
            ["Vap", "Liq"], initialize=0, units=pyunits.mol / (pyunits.m**3)
        )
        m.state[0].dens_mol_phase["Liq"].value = 3
        m.state[0].pressure.value = 5

        assert approx(split, 1e-6) == value(
            henry_equilibrium_ratio(m.state[0], "Liq", "H2O")
        )
        assert (
            str(pyunits.get_units(henry_equilibrium_ratio(m.state[0], "Liq", "H2O")))
            == "dimensionless"
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from .feedwater_heater_0D import FWH0D, FWHCondensing0D
from .balance import BalanceBlockData, BalanceBlock
from .boiler_fireside import BoilerFireside
from .boiler_heat_exchanger import BoilerHeatExchanger
from .boiler_heat_exchanger_2D import HeatExchangerCrossFlow2D_Header
from .cross_flow_heat_exchanger_1D import (
    CrossFlowHeatExchanger1D,
    CrossFlowHeatExchanger1DInitializer,
)
from .downcomer import Downcomer
from .drum import Drum
from .drum1D import Drum1D
from .feedwater_heater_0D_dynamic import FWH0DDynamic
from .heater_1D import Heater1D, Heater1DInitializer
from .heat_exchanger_3streams import HeatExchangerWith3Streams
from .steamheater import SteamHeater
from .waterpipe import WaterPipe
from .watertank import WaterTank
from .waterwall_section import WaterwallSection
